from cog import BasePredictor, Path


class Predictor(BasePredictor):
    def predict(self, path: Path) -> str:
        with open(path) as fh:
            extension = fh.name.split(".")[-1]
            return f"{extension} {fh.read()}"



class Predictor:
    async def download(self) -> None:
        print("setup used asyncio.run! it's not very effective...")

    async def setup(self) -> None:
        await self.download()

    def predict(self) -> str:
        return "output"



import asyncio
import os
import pathlib
import shutil
import subprocess
import time
from pathlib import Path

import httpx
import pytest

from .util import cog_server_http_run

DEFAULT_TIMEOUT = 60


def test_predict_takes_string_inputs_and_returns_strings_to_stdout():
    project_dir = Path(__file__).parent / "fixtures/string-project"
    result = subprocess.run(
        ["cog", "predict", "--debug", "-i", "s=world"],
        cwd=project_dir,
        check=True,
        capture_output=True,
        text=True,
        timeout=DEFAULT_TIMEOUT,
    )
    # stdout should be clean without any log messages so it can be piped to other commands
    assert result.stdout == "hello world\n"
    assert "cannot use fast loader as current Python <3.9" in result.stderr
    assert "falling back to slow loader" in result.stderr


def test_predict_supports_async_predictors():
    project_dir = Path(__file__).parent / "fixtures/async-string-project"
    result = subprocess.run(
        ["cog", "predict", "--debug", "-i", "s=world"],
        cwd=project_dir,
        check=True,
        capture_output=True,
        text=True,
        timeout=DEFAULT_TIMEOUT,
    )
    # stdout should be clean without any log messages so it can be piped to other commands
    assert result.stdout == "hello world\n"


def test_predict_takes_int_inputs_and_returns_ints_to_stdout():
    project_dir = Path(__file__).parent / "fixtures/int-project"
    result = subprocess.run(
        ["cog", "predict", "--debug", "-i", "num=2"],
        cwd=project_dir,
        check=True,
        capture_output=True,
        text=True,
        timeout=DEFAULT_TIMEOUT,
    )
    # stdout should be clean without any log messages so it can be piped to other commands
    assert result.stdout == "4\n"
    assert "falling back to slow loader" not in result.stderr


def test_predict_takes_file_inputs(tmpdir_factory):
    project_dir = Path(__file__).parent / "fixtures/path-input-project"
    out_dir = pathlib.Path(tmpdir_factory.mktemp("project"))
    shutil.copytree(project_dir, out_dir, dirs_exist_ok=True)
    with open(out_dir / "input.txt", "w", encoding="utf-8") as fh:
        fh.write("what up")
    result = subprocess.run(
        ["cog", "predict", "--debug", "-i", "path=@" + str(out_dir / "input.txt")],
        cwd=out_dir,
        check=True,
        capture_output=True,
        text=True,
        timeout=DEFAULT_TIMEOUT,
    )
    assert result.stdout == "what up\n"
    assert "falling back to slow loader" not in result.stderr


def test_predict_writes_files_to_files(tmpdir_factory):
    project_dir = Path(__file__).parent / "fixtures/path-output-project"
    out_dir = pathlib.Path(tmpdir_factory.mktemp("project"))
    shutil.copytree(project_dir, out_dir, dirs_exist_ok=True)
    result = subprocess.run(
        ["cog", "predict", "--debug"],
        cwd=out_dir,
        check=True,
        capture_output=True,
        text=True,
        timeout=DEFAULT_TIMEOUT,
    )
    assert result.stdout == ""
    with open(out_dir / "output.bmp", "rb") as f:
        assert len(f.read()) == 195894
    assert "falling back to slow loader" not in result.stderr


def test_predict_writes_files_to_files_with_custom_name(tmpdir_factory):
    project_dir = Path(__file__).parent / "fixtures/path-output-project"
    out_dir = pathlib.Path(tmpdir_factory.mktemp("project"))
    shutil.copytree(project_dir, out_dir, dirs_exist_ok=True)
    result = subprocess.run(
        ["cog", "predict", "--debug", "-o", out_dir / "myoutput.bmp"],
        cwd=out_dir,
        check=True,
        capture_output=True,
        text=True,
        timeout=DEFAULT_TIMEOUT,
    )
    assert result.stdout == ""
    with open(out_dir / "myoutput.bmp", "rb") as f:
        assert len(f.read()) == 195894
    assert "falling back to slow loader" not in result.stderr


def test_predict_writes_multiple_files_to_files(tmpdir_factory):
    project_dir = Path(__file__).parent / "fixtures/path-list-output-project"

    out_dir = pathlib.Path(tmpdir_factory.mktemp("project"))
    shutil.copytree(project_dir, out_dir, dirs_exist_ok=True)

    result = subprocess.run(
        ["cog", "predict"],
        cwd=out_dir,
        check=True,
        capture_output=True,
        text=True,
        timeout=DEFAULT_TIMEOUT,
    )

    assert result.stdout == ""
    with open(out_dir / "output.0.txt", encoding="utf-8") as f:
        assert f.read() == "foo"
    with open(out_dir / "output.1.txt", encoding="utf-8") as f:
        assert f.read() == "bar"
    with open(out_dir / "output.2.txt", encoding="utf-8") as f:
        assert f.read() == "baz"
    assert "falling back to slow loader" not in result.stderr


def test_predict_writes_strings_to_files(tmpdir_factory):
    project_dir = Path(__file__).parent / "fixtures/string-project"
    out_dir = pathlib.Path(tmpdir_factory.mktemp("project"))
    result = subprocess.run(
        ["cog", "predict", "--debug", "-i", "s=world", "-o", out_dir / "out.txt"],
        cwd=project_dir,
        check=True,
        capture_output=True,
        text=True,
        timeout=DEFAULT_TIMEOUT,
    )
    assert result.stdout == ""
    with open(out_dir / "out.txt", encoding="utf-8") as f:
        assert f.read() == "hello world"
    assert "cannot use fast loader as current Python <3.9" in result.stderr
    assert "falling back to slow loader" in result.stderr


def test_predict_runs_an_existing_image(docker_image, tmpdir_factory):
    project_dir = Path(__file__).parent / "fixtures/string-project"

    subprocess.run(
        ["cog", "build", "--debug", "-t", docker_image],
        cwd=project_dir,
        check=True,
    )

    # Run in another directory to ensure it doesn't use cog.yaml
    another_directory = tmpdir_factory.mktemp("project")
    result = subprocess.run(
        ["cog", "predict", "--debug", docker_image, "-i", "s=world"],
        cwd=another_directory,
        capture_output=True,
        text=True,
        timeout=DEFAULT_TIMEOUT,
    )
    assert result.returncode == 0
    assert result.stdout == "hello world\n"
    assert "cannot use fast loader as current Python <3.9" in result.stderr
    assert "falling back to slow loader" in result.stderr


# https://github.com/replicate/cog/commit/28202b12ea40f71d791e840b97a51164e7be3b3c
# we need to find a better way to test this
@pytest.mark.skip("incredibly slow")
def test_predict_with_remote_image(tmpdir_factory):
    image_name = "r8.im/replicate/hello-world@sha256:5c7d5dc6dd8bf75c1acaa8565735e7986bc5b66206b55cca93cb72c9bf15ccaa"
    subprocess.run(["docker", "rmi", "-f", image_name], check=True)

    # Run in another directory to ensure it doesn't use cog.yaml
    another_directory = tmpdir_factory.mktemp("project")
    result = subprocess.run(
        ["cog", "predict", image_name, "-i", "text=world"],
        cwd=another_directory,
        check=True,
        capture_output=True,
        text=True,
        timeout=DEFAULT_TIMEOUT,
    )

    # lots of docker pull logs are written to stdout before writing the actual output
    # TODO: clean up docker output so cog predict is always clean
    assert result.stdout.strip().endswith("hello world")


def test_predict_in_subdirectory_with_imports(tmpdir_factory):
    project_dir = Path(__file__).parent / "fixtures/subdirectory-project"
    result = subprocess.run(
        ["cog", "predict", "--debug", "-i", "s=world"],
        cwd=project_dir,
        check=True,
        capture_output=True,
        text=True,
        timeout=DEFAULT_TIMEOUT,
    )
    # stdout should be clean without any log messages so it can be piped to other commands
    assert result.stdout == "hello world\n"
    assert "falling back to slow loader" not in result.stderr


def test_predict_many_inputs(tmpdir_factory):
    project_dir = Path(__file__).parent / "fixtures/many-inputs-project"
    out_dir = pathlib.Path(tmpdir_factory.mktemp("project"))
    shutil.copytree(project_dir, out_dir, dirs_exist_ok=True)
    inputs = {
        "no_default": "hello",
        "path": "@path.txt",
        "image": "@image.jpg",
        "choices": "foo",
        "int_choices": 3,
    }
    with open(out_dir / "path.txt", "w", encoding="utf-8") as fh:
        fh.write("world")
    with open(out_dir / "image.jpg", "w", encoding="utf-8") as fh:
        fh.write("")
    cmd = ["cog", "--debug", "predict"]

    for k, v in inputs.items():
        cmd += ["-i", f"{k}={v}"]

    result = subprocess.run(
        cmd,
        cwd=out_dir,
        check=True,
        capture_output=True,
        text=True,
        timeout=DEFAULT_TIMEOUT,
    )
    assert result.stdout == "hello default 20 world jpg foo 6\n"
    assert "falling back to slow loader" not in result.stderr


def test_predict_many_inputs_with_existing_image(docker_image, tmpdir_factory):
    project_dir = Path(__file__).parent / "fixtures/many-inputs-project"

    subprocess.run(
        ["cog", "build", "--debug", "-t", docker_image],
        cwd=project_dir,
        check=True,
    )

    out_dir = pathlib.Path(tmpdir_factory.mktemp("project"))
    shutil.copytree(project_dir, out_dir, dirs_exist_ok=True)
    inputs = {
        "no_default": "hello",
        "path": "@path.txt",
        "image": "@image.jpg",
        "choices": "foo",
        "int_choices": 3,
    }
    with open(out_dir / "path.txt", "w", encoding="utf-8") as fh:
        fh.write("world")
    with open(out_dir / "image.jpg", "w", encoding="utf-8") as fh:
        fh.write("")
    cmd = ["cog", "--debug", "predict", docker_image]

    for k, v in inputs.items():
        cmd += ["-i", f"{k}={v}"]

    result = subprocess.run(
        cmd,
        cwd=out_dir,
        check=True,
        capture_output=True,
        text=True,
    )
    assert result.stdout == "hello default 20 world jpg foo 6\n"
    assert "falling back to slow loader" not in str(result.stderr)


def test_predict_path_list_input(tmpdir_factory):
    project_dir = Path(__file__).parent / "fixtures/path-list-input-project"
    out_dir = pathlib.Path(tmpdir_factory.mktemp("project"))
    shutil.copytree(project_dir, out_dir, dirs_exist_ok=True)
    with open(out_dir / "1.txt", "w", encoding="utf-8") as fh:
        fh.write("test1")
    with open(out_dir / "2.txt", "w", encoding="utf-8") as fh:
        fh.write("test2")
    cmd = ["cog", "predict", "-i", "paths=@1.txt", "-i", "paths=@2.txt"]

    result = subprocess.run(
        cmd,
        cwd=out_dir,
        check=True,
        capture_output=True,
        text=True,
        timeout=DEFAULT_TIMEOUT,
    )
    assert "test1" in result.stdout
    assert "test2" in result.stdout


@pytest.mark.parametrize(
    ("fixture_name",),
    [
        ("simple",),
        ("double-fork",),
        ("double-fork-http",),
        ("multiprocessing",),
    ],
)
def test_predict_with_subprocess_in_setup(fixture_name):
    project_dir = (
        Path(__file__).parent / "fixtures" / f"setup-subprocess-{fixture_name}-project"
    )

    with cog_server_http_run(project_dir) as addr:
        busy_count = 0

        for i in range(100):
            response = httpx.post(
                f"{addr}/predictions",
                json={"input": {"s": f"friendo{i}"}},
            )
            if response.status_code == 409:
                busy_count += 1
                continue

            assert response.status_code == 200, str(response)

        assert busy_count < 10


@pytest.mark.asyncio
async def test_concurrent_predictions():
    async def make_request(i: int) -> httpx.Response:
        return await client.post(
            f"{addr}/predictions",
            json={
                "id": f"id-{i}",
                "input": {"s": f"sleepyhead{i}", "sleep": 1.0},
            },
        )

    with cog_server_http_run(
        Path(__file__).parent / "fixtures" / "async-sleep-project"
    ) as addr:
        async with httpx.AsyncClient() as client:
            tasks = []
            start = time.perf_counter()
            async with asyncio.TaskGroup() as tg:
                for i in range(5):
                    tasks.append(tg.create_task(make_request(i)))
                # give time for all of the predictions to be accepted, but not completed
                await asyncio.sleep(0.2)
                # we shut the server down, but expect all running predictions to complete
                await client.post(f"{addr}/shutdown")
            end = time.perf_counter()
            assert (end - start) < 3.0  # ensure the predictions ran concurrently
            for i, task in enumerate(tasks):
                assert task.result().status_code == 200
                assert task.result().json()["output"] == f"wake up sleepyhead{i}"


def test_predict_new_union_project(tmpdir_factory):
    project_dir = Path(__file__).parent / "fixtures/new-union-project"
    result = subprocess.run(
        ["cog", "predict", "--debug", "-i", "text=world"],
        cwd=project_dir,
        check=True,
        capture_output=True,
        text=True,
        timeout=DEFAULT_TIMEOUT,
    )
    # stdout should be clean without any log messages so it can be piped to other commands
    assert result.returncode == 0
    assert result.stdout == "hello world\n"


def test_predict_with_fast_build_with_local_image(docker_image):
    project_dir = Path(__file__).parent / "fixtures/fast-build"
    weights_file = os.path.join(project_dir, "weights.h5")
    with open(weights_file, "w", encoding="utf8") as handle:
        handle.seek(256 * 1024 * 1024)
        handle.write("\0")

    build_process = subprocess.run(
        ["cog", "build", "-t", docker_image, "--x-fast", "--x-localimage"],
        cwd=project_dir,
        capture_output=True,
    )

    result = subprocess.run(
        [
            "cog",
            "predict",
            docker_image,
            "--x-fast",
            "--x-localimage",
            "--debug",
            "-i",
            "s=world",
        ],
        cwd=project_dir,
        capture_output=True,
    )
    os.remove(weights_file)
    assert build_process.returncode == 0
    assert result.returncode == 0


def test_predict_optional_project(tmpdir_factory):
    project_dir = Path(__file__).parent / "fixtures/optional-project"
    result = subprocess.run(
        ["cog", "predict", "--debug"],
        cwd=project_dir,
        check=True,
        capture_output=True,
        text=True,
        timeout=DEFAULT_TIMEOUT,
    )
    # stdout should be clean without any log messages so it can be piped to other commands
    assert result.returncode == 0
    assert result.stdout == "hello No One\n"


def test_predict_complex_types(docker_image):
    project_dir = Path(__file__).parent / "fixtures/complex-types"

    build_process = subprocess.run(
        ["cog", "build", "-t", docker_image, "--x-fast", "--x-localimage"],
        cwd=project_dir,
        capture_output=True,
    )
    assert build_process.returncode == 0
    result = subprocess.run(
        [
            "cog",
            "predict",
            "--debug",
            docker_image,
            "-i",
            'message={"content": "Hi There", "role": "user"}',
        ],
        cwd=project_dir,
        check=True,
        capture_output=True,
        text=True,
        timeout=DEFAULT_TIMEOUT,
    )
    assert result.returncode == 0
    assert result.stdout == "Content: Hi There\n"


def test_predict_overrides_project(docker_image):
    project_dir = Path(__file__).parent / "fixtures/overrides-project"
    build_process = subprocess.run(
        ["cog", "build", "-t", docker_image],
        cwd=project_dir,
        capture_output=True,
    )
    assert build_process.returncode == 0
    result = subprocess.run(
        ["cog", "predict", "--debug", docker_image],
        cwd=project_dir,
        check=True,
        capture_output=True,
        text=True,
        timeout=DEFAULT_TIMEOUT,
    )
    assert result.returncode == 0
    assert result.stdout == "hello 1.26.4\n"


def test_predict_zsh_package(docker_image):
    project_dir = Path(__file__).parent / "fixtures/zsh-package"
    build_process = subprocess.run(
        ["cog", "build", "-t", docker_image],
        cwd=project_dir,
        capture_output=True,
    )
    assert build_process.returncode == 0
    result = subprocess.run(
        ["cog", "predict", "--debug", docker_image],
        cwd=project_dir,
        check=True,
        capture_output=True,
        text=True,
        timeout=DEFAULT_TIMEOUT,
    )
    assert result.returncode == 0
    assert ",sh," in result.stdout
    assert ",zsh," in result.stdout



import os
import resource
import time
import uuid

import pytest

from cog.server.helpers import StreamRedirector


@pytest.fixture
def tmpfile(tmp_path):
    def _tmpfile():
        return tmp_path / uuid.uuid4().hex

    return _tmpfile


def test_stream_redirector_multiple_streams(tmpfile):
    stdout_file = tmpfile()
    stderr_file = tmpfile()
    fake_stdout = stdout_file.open("w")
    fake_stderr = stderr_file.open("w")
    output = []

    def _write_hook(stream_name, data):
        output.append((stream_name, data))

    r = StreamRedirector(callback=_write_hook, streams=[fake_stdout, fake_stderr])

    with r:
        fake_stdout.write("hello to stdout\n")
        fake_stdout.flush()
        fake_stderr.write("hello to stderr\n")
        fake_stderr.flush()

        r.drain()

    assert stdout_file.read_text() == ""
    assert stderr_file.read_text() == ""
    assert output == [
        (stdout_file.as_posix(), "hello to stdout\n"),
        (stderr_file.as_posix(), "hello to stderr\n"),
    ]


def test_stream_redirector_bench():
    fake_stdout = open(os.devnull, "w")

    def _write_hook(stream_name, data):
        pass

    r = StreamRedirector(callback=_write_hook, streams=[fake_stdout])

    n = 1024
    i = 0
    now = time.monotonic()

    with r:
        while time.monotonic() - now < 1:
            fake_stdout.write("0" * (n - 1) + "\n")
            fake_stdout.flush()
            i += 1
        r.drain()

    delta = time.monotonic() - now

    print(
        f"wrote {i} lines in {delta:.4f} seconds ({i / delta:.2f} lines/sec, {i * n / delta:.2f} bytes/sec)"
    )
    # We should be able to push at least 100MB/s through the stream redirector.
    #
    # For reasons we're not yet sure of, throughput on GitHub Actions is really
    # poor, so we're setting the threshold to 25MB/s for now.
    assert i * n / delta > 25e6

    fake_stdout.close()


@pytest.mark.parametrize(
    "writes,expected_output",
    [
        # lines are preserved
        (["hello world\n"], ["hello world\n"]),
        # partial lines at end are flushed (complete with a newline that wasn't
        # actually written)
        (["one\n", "two"], ["one\n", "two\n"]),
        # partial lines in the middle are buffered until a newline is seen
        (["one", "two", "three\n"], ["onetwothree\n"]),
    ],
)
def test_stream_redirector_line_handling(tmpfile, writes, expected_output):
    f = tmpfile()
    stream = f.open("w")
    output = []

    def _write_hook(stream_name, data):
        output.append((stream_name, data))

    r = StreamRedirector(callback=_write_hook, streams=[stream])

    with r:
        for w in writes:
            stream.write(w)
            stream.flush()

        r.drain()

    stream.write("no longer redirected\n")
    stream.flush()

    assert f.read_text() == "no longer redirected\n"
    assert output == [(f.as_posix(), o) for o in expected_output]

    stream.close()


def test_stream_redirector_reentrant(tmpfile):
    f = tmpfile()
    stream = f.open("w")
    output = []

    def _write_hook(stream_name, data):
        output.append((stream_name, data))

    r = StreamRedirector(callback=_write_hook, streams=[stream])

    with r:
        stream.write("one\n")
        stream.flush()

        with r:
            stream.write("two\n")
            stream.flush()

            with r:
                stream.write("three\n")
                stream.flush()

        r.drain()

    stream.write("four\n")
    stream.flush()
    stream.close()

    assert f.read_text() == "four\n"
    assert output == [
        (f.as_posix(), "one\n"),
        (f.as_posix(), "two\n"),
        (f.as_posix(), "three\n"),
    ]


def test_stream_redirector_tee(tmpfile):
    f = tmpfile()
    stream = f.open("w")
    output = []

    def _write_hook(stream_name, data):
        output.append((stream_name, data))

    r = StreamRedirector(callback=_write_hook, tee=True, streams=[stream])

    with r:
        stream.write("one\n")
        stream.write("two\n")
        stream.write("three\n")
        stream.flush()

        r.drain()

    stream.write("four\n")
    stream.flush()
    stream.close()

    assert f.read_text() == "one\ntwo\nthree\nfour\n"
    assert output == [
        (f.as_posix(), "one\n"),
        (f.as_posix(), "two\n"),
        (f.as_posix(), "three\n"),
    ]


def test_stream_redirector_does_not_leak_file_descriptors(tmpfile, request):
    f = tmpfile()
    stream = f.open("w")

    def _write_hook(stream_name, data):
        pass

    original_limits = resource.getrlimit(resource.RLIMIT_NOFILE)
    resource.setrlimit(resource.RLIMIT_NOFILE, (256, original_limits[1]))
    request.addfinalizer(
        lambda: resource.setrlimit(resource.RLIMIT_NOFILE, original_limits)
    )

    r = StreamRedirector(callback=_write_hook, streams=[stream])

    for _ in range(10 * 256):
        with r:
            stream.write("one\n")
            stream.flush()
            r.drain()

    stream.close()



from cog import BasePredictor, Input, Path


class Predictor(BasePredictor):
    def predict(
        self,
        no_default: str,
        default_without_input: str = "default",
        input_with_default: int = Input(default=10),
        path: Path = Input(description="Some path"),
        image: Path = Input(description="Some path"),
        choices: str = Input(choices=["foo", "bar"]),
        int_choices: int = Input(description="hello", choices=[3, 4, 5]),
    ) -> str:
        with path.open() as f:
            path_contents = f.read()
        image_extension = str(image).split(".")[-1]
        return (
            no_default
            + " "
            + default_without_input
            + " "
            + str(input_with_default * 2)
            + " "
            + path_contents
            + " "
            + image_extension
            + " "
            + choices
            + " "
            + str(int_choices * 2)
        )



import sys
import typing

if typing.TYPE_CHECKING:
    from typing_extensions import Protocol
else:
    Protocol = object


class IMimeTypes(Protocol):
    def add_type(self, type: str, ext: str, strict: bool = True) -> None: ...


def install_mime_extensions(mimetypes: IMimeTypes) -> None:
    """
    Older versions of Python are missing the MIME types for more recent file formats
    this function adds the missing MIME types to the mimetypes module.
    """

    # This could also be done by loading a mime.types file from disk using
    # mimetypes.read_mime_types().
    if sys.version_info < (3, 13):
        mimetypes.add_type("image/webp", ".webp")



import ctypes
import logging
import sys
import time

libc = ctypes.CDLL(None)

# test that we can still capture type signature even if we write
# a bunch of stuff at import time.
libc.puts(b"writing some stuff from C at import time")
libc.fflush(None)
sys.stdout.write("writing to stdout at import time\n")
sys.stderr.write("writing to stderr at import time\n")


class Predictor:
    def setup(self):
        print("setting up predictor")
        self.foo = "foo"

    async def predict(self) -> str:
        time.sleep(0.1)
        logging.warn("writing log message")
        time.sleep(0.1)
        libc.puts(b"writing from C")  # not expected to be seen
        libc.fflush(None)
        time.sleep(0.1)
        sys.stderr.write("writing to stderr\n")
        time.sleep(0.1)
        sys.stderr.flush()
        time.sleep(0.1)
        print("writing with print")
        time.sleep(0.1)
        return "output"



import builtins
import enum
import importlib.util
import inspect
import os.path
import sys
import types
import uuid
from collections.abc import Iterable, Iterator

if sys.version_info >= (3, 10):
    from types import NoneType
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Literal,
    Optional,
    Type,
    Union,
    cast,
    get_args,
    get_origin,
)
from unittest.mock import patch

import pydantic
import structlog
from pydantic import BaseModel, Field, create_model
from pydantic.fields import FieldInfo

# Added in Python 3.9. Can be from typing if we drop support for <3.9
from typing_extensions import Annotated

from .base_input import BaseInput
from .base_predictor import BasePredictor
from .code_xforms import load_module_from_string, strip_model_source_code
from .types import (
    PYDANTIC_V2,
    Input,
    Weights,
)
from .types import (
    File as CogFile,
)
from .types import (
    Path as CogPath,
)
from .types import Secret as CogSecret

if PYDANTIC_V2:
    from pydantic.fields import PydanticUndefined  # type: ignore
else:
    from pydantic.fields import Undefined as PydanticUndefined

log = structlog.get_logger("cog.server.predictor")

ALLOWED_INPUT_TYPES: List[Type[Any]] = [
    str,
    int,
    float,
    bool,
    CogFile,
    CogPath,
    CogSecret,
]


def has_setup_weights(predictor: BasePredictor) -> bool:
    weights_type = get_weights_type(predictor.setup)
    return weights_type is not None


def extract_setup_weights(predictor: BasePredictor) -> Optional[Weights]:
    weights_type = get_weights_type(predictor.setup)
    assert weights_type

    weights: Optional[Weights]

    weights_url = os.environ.get("COG_WEIGHTS")
    weights_path = "weights"

    # TODO: Cog{File,Path}.validate(...) methods accept either "real"
    # paths/files or URLs to those things. In future we can probably tidy this
    # up a little bit.
    # TODO: CogFile/CogPath should have subclasses for each of the subtypes
    if weights_url:
        if PYDANTIC_V2:
            from pydantic import TypeAdapter

            for t in [CogFile, CogPath]:
                try:
                    weights = TypeAdapter(t).validate_python(weights_url)
                    break
                except Exception:  # pylint: disable=broad-except # noqa: S110
                    pass
            else:
                if weights_type is str:
                    weights = weights_url
                else:
                    raise ValueError(
                        f"Predictor.setup() has an argument 'weights' of type {weights_type}, but only File, Path and str are supported"
                    )
        else:
            if weights_type is CogFile:
                weights = cast(CogFile, CogFile.validate(weights_url))
            elif weights_type is CogPath:
                # TODO: So this can be a url. evil!
                weights = cast(CogPath, CogPath.validate(weights_url))
            elif weights_type is str:
                weights = weights_url
            else:
                raise ValueError(
                    f"Predictor.setup() has an argument 'weights' of type {weights_type}, but only File, Path and str are supported"
                )
    elif os.path.exists(weights_path):
        if weights_type == CogFile:
            with open(weights_path, "rb") as f:
                weights = cast(CogFile, f)
        elif weights_type == CogPath:
            weights = CogPath(weights_path)
        else:
            raise ValueError(
                f"Predictor.setup() has an argument 'weights' of type {weights_type}, but only File, Path and str are supported"
            )
    else:
        weights = None

    return weights


def get_weights_type(setup_function: Callable[[Any], None]) -> Optional[Any]:
    signature = inspect.signature(setup_function)
    if "weights" not in signature.parameters:
        return None
    Type = signature.parameters["weights"].annotation  # pylint: disable=invalid-name,redefined-outer-name
    # Handle Optional. It is Union[Type, None]
    if get_origin(Type) == Union:
        args = get_args(Type)
        if len(args) == 2 and args[1] is type(None):
            Type = get_args(Type)[0]  # pylint: disable=invalid-name
    return Type


def load_full_predictor_from_file(
    module_path: str, module_name: str
) -> types.ModuleType:
    spec = importlib.util.spec_from_file_location(module_name, module_path)
    assert spec is not None
    module = importlib.util.module_from_spec(spec)
    assert spec.loader is not None
    # Remove any sys.argv while importing predictor to avoid conflicts when
    # user code calls argparse.Parser.parse_args in production
    with patch("sys.argv", sys.argv[:1]):
        spec.loader.exec_module(module)
    return module


def load_slim_predictor_from_file(
    module_path: str, class_name: str, method_name: str
) -> Optional[types.ModuleType]:
    with open(module_path, encoding="utf-8") as file:
        source_code = file.read()
    stripped_source = strip_model_source_code(source_code, [class_name], [method_name])
    module = load_module_from_string(uuid.uuid4().hex, stripped_source)
    return module


def get_predictor(module: types.ModuleType, class_name: str) -> Any:
    predictor = getattr(module, class_name)
    # It could be a class or a function
    if inspect.isclass(predictor):
        return predictor()
    return predictor


def load_predictor_from_ref(ref: str) -> BasePredictor:
    module_path, class_name = ref.split(":", 1)
    module_name = os.path.basename(module_path).split(".py", 1)[0]
    module = load_full_predictor_from_file(module_path, module_name)
    predictor = get_predictor(module, class_name)
    return predictor


def is_union(type: Type[Any]) -> bool:
    if get_origin(type) is Union:
        return True
    if hasattr(types, "UnionType") and get_origin(type) is types.UnionType:
        return True
    return False


def is_optional(type: Type[Any]) -> bool:
    args = get_args(type)
    if len(args) != 2 or not is_union(type):
        return False
    if sys.version_info >= (3, 10):
        return args[1] is NoneType
    return args[1] is None.__class__


def validate_input_type(
    type: Type[Any],  # pylint: disable=redefined-builtin
    name: str,
) -> None:
    if type is inspect.Signature.empty:
        raise TypeError(
            f"No input type provided for parameter `{name}`. Supported input types are: {readable_types_list(ALLOWED_INPUT_TYPES)}, or a Union or List of those types."
        )
    if type not in ALLOWED_INPUT_TYPES:
        if get_origin(type) is Literal:
            for t in get_args(type):
                validate_input_type(builtins.type(t), name)
        elif get_origin(type) in (Union, List, list) or is_union(type):  # noqa: E721
            args = get_args(type)
            if is_optional(type):
                validate_input_type(args[0], name)
            else:
                for t in args:
                    validate_input_type(t, name)
        else:
            if PYDANTIC_V2:
                # Cog types are exported as `Annotated[Type, ...]`, but `type` is the inner type
                if hasattr(type, "__module__") and type.__module__ == "cog.types":
                    return

            raise TypeError(
                f"Unsupported input type {human_readable_type_name(type)} for parameter `{name}`. Supported input types are: {readable_types_list(ALLOWED_INPUT_TYPES)}, or a Union or List of those types."
            )


def get_input_create_model_kwargs(signature: inspect.Signature) -> Dict[str, Any]:
    create_model_kwargs = {}

    order = 0

    for name, parameter in signature.parameters.items():
        InputType = parameter.annotation

        validate_input_type(InputType, name)

        # if no default is specified, create an empty, required input
        if parameter.default is inspect.Signature.empty:
            default = Input()
        else:
            if not isinstance(parameter.default, FieldInfo):
                default = Input(default=parameter.default)
            else:
                if is_optional(InputType):
                    # If we are an optional, make sure the default is None
                    if (
                        parameter.default.default is PydanticUndefined
                        or parameter.default.default is ...
                    ):
                        if PYDANTIC_V2:
                            parameter.default.default = None
                        else:
                            parameter.default.default_factory = None
                            parameter.default.default = None
                default = parameter.default

        if PYDANTIC_V2:
            # https://github.com/pydantic/pydantic/blob/2.7/pydantic/json_schema.py#L1436-L1446
            # json_schema_extra can be a callable, but we don't set that and users shouldn't set that
            if not default.json_schema_extra:  # type: ignore
                default.json_schema_extra = {}  # type: ignore
            assert isinstance(default.json_schema_extra, dict)  # type: ignore
            extra = default.json_schema_extra  # type: ignore
        else:
            extra = default.extra  # type: ignore
        extra["x-order"] = order
        order += 1

        # Choices!
        choices = (
            extra.pop("choices", None)  # Pydantic v1
            or extra.pop("enum", None)  # Pydantic v2
        )
        # In either case, remove it as an extra field because it will be
        # passed automatically as 'enum' in the schema
        if choices:
            if InputType == str and isinstance(choices, Iterable):  # noqa: E721

                class StringEnum(str, enum.Enum):
                    pass

                InputType = StringEnum(  # pylint: disable=invalid-name
                    name, [(value, value) for value in choices or []]
                )
            elif InputType == int:  # noqa: E721
                InputType = enum.IntEnum(name, {str(value): value for value in choices})  # type: ignore # pylint: disable=invalid-name
            else:
                raise TypeError(
                    f"The input {name} uses the option choices. Choices can only be used with str or int types."
                )

        create_model_kwargs[name] = (InputType, default)

    return create_model_kwargs


def get_predict(predictor: Any) -> Callable[..., Any]:
    if hasattr(predictor, "predict"):
        return predictor.predict
    return predictor


def get_input_type(predictor: BasePredictor) -> Type[BaseInput]:
    """
    Creates a Pydantic Input model from the arguments of a Predictor's predict() method.

    class Predictor(BasePredictor):
        def predict(self, text: str):
            ...

    programmatically creates a model like this:

    class Input(BaseModel):
        text: str
    """

    predict = get_predict(predictor)
    signature = inspect.signature(predict)

    return create_model(
        "Input",
        __config__=None,
        __base__=BaseInput,
        __module__=__name__,
        __validators__=None,
        **get_input_create_model_kwargs(signature),
    )  # type: ignore


def get_output_type(predictor: BasePredictor) -> Type[BaseModel]:
    """
    Creates a Pydantic Output model from the return type annotation of a Predictor's predict() method.
    """

    predict = get_predict(predictor)
    signature = inspect.signature(predict)
    OutputType: Type[BaseModel]
    if signature.return_annotation is inspect.Signature.empty:
        raise TypeError(
            """You must set an output type. If your model can return multiple output types, you can explicitly set `Any` as the output type.

For example:

    from typing import Any

    def predict(
        self,
        image: Path = Input(description="Input image"),
    ) -> Any:
        ...
"""
        )
    else:
        OutputType = signature.return_annotation

    # The type that goes in the response is a list of the yielded type
    if get_origin(OutputType) is Iterator:
        # Annotated allows us to attach Field annotations to the list, which we use to mark that this is an iterator
        # https://pydantic-docs.helpmanual.io/usage/schema/#typingannotated-fields
        if PYDANTIC_V2:
            field = Field(**{"json_schema_extra": {"x-cog-array-type": "iterator"}})  # type: ignore
        else:
            field = Field(**{"x-cog-array-type": "iterator"})  # type: ignore
        OutputType: Type[BaseModel] = Annotated[List[get_args(OutputType)[0]], field]  # type: ignore

    name = OutputType.__name__ if hasattr(OutputType, "__name__") else ""

    if name == "Output":
        return OutputType

    # We wrap the OutputType in an Output class to
    # ensure consistent naming of the interface in the schema.
    #
    # NOTE: If the OutputType.__name__ is "TrainingOutput" then cannot use
    # `__root__` here because this will create a reference for the Object.
    # e.g.
    #   {'title': 'Output', '$ref': '#/definitions/TrainingOutput' ... }
    #
    # And this reference may conflict with other objects at which
    # point the item will be namespaced and break our parsing. e.g.
    #   {'title': 'Output', '$ref': '#/definitions/predict_TrainingOutput' ... }
    #
    # So we work around this by inheriting from the original class rather
    # than using "__root__".
    if name == "TrainingOutput":  # pylint: disable=no-else-return

        class Output(OutputType):  # type: ignore
            pass

        return Output
    else:
        if PYDANTIC_V2:

            class Output(pydantic.RootModel[OutputType]):  # type: ignore
                pass
        else:

            class Output(BaseModel):
                __root__: OutputType  # type: ignore

        return Output


def get_train(predictor: Any) -> Callable[..., Any]:
    if hasattr(predictor, "train"):
        return predictor.train
    return predictor


def get_training_input_type(predictor: BasePredictor) -> Type[BaseInput]:
    """
    Creates a Pydantic Input model from the arguments of a Predictor's train() method.

    def train(self, text: str):
        ...

    programmatically creates a model like this:

    class TrainingInput(BaseModel):
        text: str
    """

    train = get_train(predictor)
    signature = inspect.signature(train)

    return create_model(
        "TrainingInput",
        __config__=None,
        __base__=BaseInput,
        __module__=__name__,
        __validators__=None,
        **get_input_create_model_kwargs(signature),
    )  # type: ignore


def get_training_output_type(predictor: BasePredictor) -> Type[BaseModel]:
    """
    Creates a Pydantic Output model from the return type annotation of a train() method.
    """

    train = get_train(predictor)
    signature = inspect.signature(train)

    if signature.return_annotation is inspect.Signature.empty:
        raise TypeError(
            """You must set an output type. If your model can return multiple output types, you can explicitly set `Any` as the output type.

For example:

    from typing import Any

    def train(
        self,
        n: int
    ) -> Any:
        ...
"""
        )
    else:
        TrainingOutputType = signature.return_annotation

    name = (
        TrainingOutputType.__name__ if hasattr(TrainingOutputType, "__name__") else ""
    )

    # We wrap the OutputType in a TrainingOutput class to
    # ensure consistent naming of the interface in the schema
    # See comment in get_output_type for more info.
    if name == "TrainingOutput":
        return TrainingOutputType

    if name == "Output":  # pylint: disable=no-else-return

        class TrainingOutput(TrainingOutputType):  # type: ignore
            pass

        return TrainingOutput
    else:
        if PYDANTIC_V2:

            class TrainingOutput(pydantic.RootModel[TrainingOutputType]):  # type: ignore
                pass

            return TrainingOutput

        else:

            class TrainingOutput(BaseModel):
                __root__: TrainingOutputType  # type: ignore

            return TrainingOutput


def human_readable_type_name(t: Type[Union[Any, None]]) -> str:
    """
    Generates a useful-for-humans label for a type. For builtin types, it's just the class name (eg "str" or "int"). For other types, it includes the module (eg "pathlib.Path" or "cog.File").

    The special case for Cog modules is because the type lives in `cog.types` internally, but just `cog` when included as a dependency.
    """

    if hasattr(t, "__module__"):
        module = t.__module__

        if module == "builtins":
            return t.__qualname__

        if module.split(".")[0] == "cog":
            module = "cog"

        try:
            return f"{module}.{t.__qualname__}"
        except AttributeError:
            pass

    return str(t)


def readable_types_list(type_list: List[Type[Any]]) -> str:
    return ", ".join(human_readable_type_name(t) for t in type_list)



class Predictor:
    async def download(self) -> None:
        print("download complete!")

    async def setup(self) -> None:
        print("setup starting...")
        await self.download()
        print("setup complete!")

    async def predict(self) -> str:
        print("running prediction")
        return "output"



import os
from typing import Optional
from cog import BaseModel, Input, Path as CogPath, Secret

# We return a path to our trained adapter weights
class TrainingOutput(BaseModel):
    weights: CogPath

def train(
    # Basic input
    some_input: str = Input(
        description="A basic string input to satisfy minimum requirements.",
        default="default value",
    ),
    # String input with None default (problematic)
    hf_repo_id: Optional[str] = Input(
        description="String with None default - this causes issues.",
        default=None,
    ),
    # Secret with None default (problematic)
    hf_token: Optional[Secret] = Input(
        description="Secret with None default - this also causes issues.",
        default=None,
    ),
    # String input with empty string default (works)
    working_repo_id: str = Input(
        description="String with empty string default - this works.",
        default="",
    ),
    # Secret with empty string default (works)
    working_token: Secret = Input(
        description="Secret with empty string default - this works.",
        default="",
    ),
) -> TrainingOutput:
    """
    Minimal example to demonstrate issues with Secret inputs.
    """
    print("\n=== Minimal Cog Secret Test ===")
    print(f"cog version: {os.environ.get('COG_VERSION', 'unknown')}")
    
    # Inputs with None defaults
    print("\n-- Inputs with None defaults (problematic) --")
    print(f"hf_repo_id: {hf_repo_id}")
    if hf_token:
        print(f"hf_token: [PROVIDED]")
        try:
            value = hf_token.get_secret_value()
            print("Secret access successful")
        except Exception as e:
            print(f"Error accessing secret: {e}")
    else:
        print("hf_token: None")
    
    # Inputs with empty string defaults
    print("\n-- Inputs with empty string defaults (works) --")
    print(f"working_repo_id: {working_repo_id if working_repo_id else '(empty)'}")
    if working_token and working_token.get_secret_value():
        print(f"working_token: [PROVIDED]")
        try:
            value = working_token.get_secret_value()
            print("Secret access successful")
        except Exception as e:
            print(f"Error accessing secret: {e}")
    else:
        print("working_token: (empty)")
    
    # Create a dummy output file
    output_path = "dummy_output.txt"
    with open(output_path, "w") as f:
        f.write("This is a dummy output file.")
    
    print("\n=== Test Complete ===")
    
    # Return the dummy output path
    return TrainingOutput(weights=CogPath(output_path))



import importlib
import os
import sys
import time

import structlog

COG_WAIT_FILE_ENV_VAR = "COG_WAIT_FILE"
COG_EAGER_IMPORTS_ENV_VAR = "COG_EAGER_IMPORTS"
COG_PYENV_PATH_ENV_VAR = "COG_PYENV_PATH"
PYTHONPATH_ENV_VAR = "PYTHONPATH"
PYTHON_VERSION_ENV_VAR = "R8_PYTHON_VERSION"

log = structlog.get_logger("cog.wait")


def _wait_flag_fallen() -> bool:
    wait_file = os.environ.get(COG_WAIT_FILE_ENV_VAR)
    if wait_file is None:
        return True
    return os.path.exists(wait_file)


def _insert_pythonpath() -> None:
    pyenv_path = os.environ.get(COG_PYENV_PATH_ENV_VAR)
    if pyenv_path is None:
        return
    full_module_path = os.path.join(
        pyenv_path,
        "lib",
        "python" + os.environ[PYTHON_VERSION_ENV_VAR],
        "site-packages",
    )
    if full_module_path not in sys.path:
        sys.path.append(full_module_path)
    os.environ[PYTHONPATH_ENV_VAR] = ":".join(sys.path)


def wait_for_file(timeout: float = 60.0) -> bool:
    """Wait for a file in the environment variables."""
    wait_file = os.environ.get(COG_WAIT_FILE_ENV_VAR)
    if wait_file is None:
        return True
    if os.path.exists(wait_file):
        log.info(f"Wait file found {wait_file}...")
        return True
    log.info(f"Waiting for file {wait_file}...")
    time_taken = 0.0
    while time_taken < timeout:
        sleep_time = 0.01
        time.sleep(sleep_time)
        time_taken += sleep_time
        if os.path.exists(wait_file):
            return True
    log.info(f"Waiting for file {wait_file} timed out.")
    return False


def eagerly_import_modules() -> int:
    """Wait for python to import big modules."""
    wait_imports = os.environ.get(COG_EAGER_IMPORTS_ENV_VAR)
    import_count = 0
    if wait_imports is None:
        return import_count
    log.info(f"Eagerly importing {wait_imports}.")
    for import_statement in wait_imports.split(","):
        importlib.import_module(import_statement)
        import_count += 1
    return import_count


def wait_for_env(file_timeout: float = 60.0, include_imports: bool = True) -> bool:
    """Wait for the environment to load."""
    if _wait_flag_fallen():
        _insert_pythonpath()
        return True
    if include_imports:
        eagerly_import_modules()
    waited = wait_for_file(timeout=file_timeout)
    _insert_pythonpath()
    return waited



from enum import Enum


class Mode(Enum):
    """Enumeration over the different prediction modes."""

    PREDICT = "predict"
    TRAIN = "train"

    def __str__(self) -> str:
        return str(self.value)



from importlib.metadata import version


def _get_version() -> str:
    try:
        return version("cog")
    except Exception:  # pylint: disable=broad-exception-caught
        return "unknown"


def get_user_agent() -> str:
    return f"cog-worker/{_get_version()}"



import multiprocessing
import os
import sys
import threading
import time
import uuid
from typing import TYPE_CHECKING, Any, Dict, List, Optional

import pytest
from attrs import define, evolve, field, frozen
from hypothesis import HealthCheck, given, settings
from hypothesis import strategies as st
from hypothesis.stateful import (
    Bundle,
    RuleBasedStateMachine,
    consumes,
    multiple,
    rule,
)

from cog.server.eventtypes import (
    Cancel,
    Done,
    Envelope,
    Log,
    PredictionInput,
    PredictionMetric,
    PredictionOutput,
    PredictionOutputType,
)
from cog.server.exceptions import FatalWorkerException, InvalidStateException
from cog.server.worker import Worker, _PublicEventType

from .conftest import WorkerConfig, uses_worker, uses_worker_configs

if TYPE_CHECKING:
    from concurrent.futures import Future

# Set a longer deadline on CI as the instances are a bit slower.
settings.register_profile("ci", max_examples=200, deadline=2000)
settings.register_profile("default", max_examples=50, deadline=1500)
settings.load_profile(os.getenv("HYPOTHESIS_PROFILE", "default"))

HYPOTHESIS_TEST_TIMEOUT = (
    settings().max_examples * settings().deadline
).total_seconds() + 5

ST_NAMES = st.sampled_from(["John", "Barry", "Elspeth", "Hamid", "Ronnie", "Yasmeen"])

SETUP_FATAL_FIXTURES = [
    "exc_in_setup",
    "exc_in_setup_and_predict",
    "exc_on_import",
    "exit_in_setup",
    "exit_on_import",
    "missing_predictor",
    "nonexistent_file",
]

PREDICTION_FATAL_FIXTURES = [
    "exit_in_predict",
    "killed_in_predict",
]

RUNNABLE_FIXTURES = [
    "simple",
    "exc_in_predict",
    "missing_predict",
]

METRICS_FIXTURES = [
    (
        WorkerConfig("record_metric"),
        {"name": ST_NAMES},
        {
            "foo": 123,
        },
    ),
    (
        WorkerConfig("record_metric_async", min_python=(3, 11), is_async=True),
        {"name": ST_NAMES},
        {
            "foo": 123,
        },
    ),
]

OUTPUT_FIXTURES = [
    (
        WorkerConfig("hello_world"),
        {"name": ST_NAMES},
        lambda x: f"hello, {x['name']}",
    ),
    (
        WorkerConfig("hello_world_async", min_python=(3, 11), is_async=True),
        {"name": ST_NAMES},
        lambda x: f"hello, {x['name']}",
    ),
    (
        WorkerConfig("count_up"),
        {"upto": st.integers(min_value=0, max_value=100)},
        lambda x: list(range(x["upto"])),
    ),
    (
        WorkerConfig("complex_output"),
        {},
        lambda _: {"number": 42, "text": "meaning of life"},
    ),
]

SETUP_LOGS_FIXTURES = [
    (
        WorkerConfig("logging", setup=False),
        (
            "writing some stuff from C at import time\n"
            "writing to stdout at import time\n"
            "setting up predictor\n"
        ),
        "writing to stderr at import time\n",
    ),
    (
        WorkerConfig("logging_async", setup=False, min_python=(3, 11), is_async=True),
        "writing to stdout at import time\nsetting up predictor\n",
        "writing to stderr at import time\n",
    ),
]

PREDICT_LOGS_FIXTURES = [
    (
        WorkerConfig("logging"),
        "writing from C\nwriting with print\n",
        "WARNING:root:writing log message\nwriting to stderr\n",
    ),
    (
        WorkerConfig("logging_async", min_python=(3, 11), is_async=True),
        "writing with print\n",
        "WARNING:root:writing log message\nwriting to stderr\n",
    ),
]

SLEEP_FIXTURES = [
    WorkerConfig("sleep"),
    WorkerConfig("sleep_async", min_python=(3, 11), is_async=True),
    WorkerConfig(
        "sleep_async",
        min_python=(3, 11),
        is_async=True,
        max_concurrency=10,
    ),
]

SLEEP_NO_SETUP_FIXTURES = [
    WorkerConfig("sleep", setup=False),
    WorkerConfig("sleep_async", min_python=(3, 11), setup=False, is_async=True),
    WorkerConfig(
        "sleep_async",
        min_python=(3, 11),
        setup=False,
        is_async=True,
        max_concurrency=10,
    ),
]


@define
class Result:
    stdout_lines: List[str] = field(factory=list)
    stderr_lines: List[str] = field(factory=list)
    heartbeat_count: int = 0
    metrics: Optional[Dict[str, Any]] = None
    output_type: Optional[PredictionOutputType] = None
    output: Any = None
    done: Optional[Done] = None
    exception: Optional[Exception] = None

    event_seen: threading.Event = field(factory=threading.Event)

    @property
    def stdout(self):
        return "".join(self.stdout_lines)

    @property
    def stderr(self):
        return "".join(self.stderr_lines)

    def handle_event(self, event: _PublicEventType):
        if isinstance(event, Log) and event.source == "stdout":
            self.stdout_lines.append(event.message)
        elif isinstance(event, Log) and event.source == "stderr":
            self.stderr_lines.append(event.message)
        elif isinstance(event, Done):
            assert not self.done
            self.done = event
        elif isinstance(event, PredictionMetric):
            if self.metrics is None:
                self.metrics = {}
            self.metrics[event.name] = event.value
        elif isinstance(event, PredictionOutput):
            assert self.output_type, "Should get output type before any output"
            if self.output_type.multi:
                self.output.append(event.payload)
            else:
                assert self.output is None, (
                    "Should not get multiple outputs for output type single"
                )
                self.output = event.payload
        elif isinstance(event, PredictionOutputType):
            assert self.output_type is None, (
                "Should not get multiple output type events"
            )
            self.output_type = event
            if self.output_type.multi:
                self.output = []
        else:
            pytest.fail(f"saw unexpected event: {event}")
        self.event_seen.set()


def _process(worker, work, swallow_exceptions=False, tag=None):
    """
    Helper function to collect events generated by Worker during tests.
    """
    result = Result()
    subid = worker.subscribe(result.handle_event, tag=tag)
    try:
        work().result()
    except Exception as exc:
        result.exception = exc
        if not swallow_exceptions:
            raise
    finally:
        worker.unsubscribe(subid)
    return result


@uses_worker(SETUP_FATAL_FIXTURES, setup=False)
def test_fatalworkerexception_from_setup_failures(worker):
    """
    Any failure during setup is fatal and should raise FatalWorkerException.
    """
    with pytest.raises(FatalWorkerException):
        _process(worker, worker.setup)


@uses_worker(PREDICTION_FATAL_FIXTURES)
def test_fatalworkerexception_from_irrecoverable_failures(worker):
    """
    Certain kinds of failure during predict (crashes, unexpected exits) are
    irrecoverable and should raise FatalWorkerException.
    """
    with pytest.raises(FatalWorkerException):
        _process(worker, lambda: worker.predict({}))

    with pytest.raises(InvalidStateException):
        _process(worker, lambda: worker.predict({}))


@uses_worker(RUNNABLE_FIXTURES)
def test_no_exceptions_from_recoverable_failures(worker):
    """
    Well-behaved predictors, or those that only throw exceptions, should not
    raise.
    """
    for _ in range(5):
        _process(worker, lambda: worker.predict({}))


@uses_worker_configs(
    [
        WorkerConfig("simple"),
        WorkerConfig("simple_async", min_python=(3, 11), is_async=True),
    ]
)
def test_can_subscribe_for_a_specific_tag(worker):
    tag = "123"

    result = Result()
    subid = worker.subscribe(result.handle_event, tag=tag)

    try:
        worker.predict({}, tag="not-my-tag").result()
        assert not result.done

        worker.predict({}, tag=tag).result()
        assert result.done
        assert not result.done.canceled
        assert not result.exception
        assert result.stdout == "did predict\n"
        assert result.output == "prediction output"

    finally:
        worker.unsubscribe(subid)


@uses_worker("sleep_async", max_concurrency=5, min_python=(3, 11), is_async=True)
def test_can_run_predictions_concurrently_on_async_predictor(worker):
    subids = []

    try:
        start = time.perf_counter()
        futures = []
        results = []
        for i in range(5):
            result = Result()
            results.append(result)
            tag = f"tag-{i}"
            subids.append(worker.subscribe(result.handle_event, tag=tag))
            futures.append(worker.predict({"sleep": 0.5}, tag=tag))
            assert not result.done

        for fut in futures:
            fut.result()

        end = time.perf_counter()

        duration = end - start
        # we should take at least 0.5 seconds (the time for 1 prediction) but
        # not more than double that
        assert duration >= 0.5
        assert duration <= 1.0

        for result in results:
            assert result.done
            assert not result.done.canceled
            assert not result.exception
            assert result.stdout == "starting\n"
            assert result.output == "done in 0.5 seconds"

    finally:
        for subid in subids:
            worker.unsubscribe(subid)


@pytest.mark.skipif(
    sys.version_info >= (3, 11), reason="Testing error message on python versions <3.11"
)
@uses_worker("simple_async", setup=False, is_async=True)
def test_async_predictor_on_python_3_10_or_older_raises_error(worker):
    fut = worker.setup()
    result = Result()
    worker.subscribe(result.handle_event)

    with pytest.raises(FatalWorkerException):
        fut.result()
    assert result.done
    assert result.done.error
    assert (
        result.done.error_detail
        == "Cog requires Python >=3.11 for `async def predict()` support"
    )


@uses_worker(
    "setup_async", max_concurrency=1, min_python=(3, 11), is_async=True, setup=False
)
def test_setup_async(worker: Worker):
    fut = worker.setup()
    setup_result = Result()
    setup_sid = worker.subscribe(setup_result.handle_event)

    # with pytest.raises(FatalWorkerException):
    fut.result()
    worker.unsubscribe(setup_sid)

    assert setup_result.stdout_lines == [
        "setup starting...\n",
        "download complete!\n",
        "setup complete!\n",
    ]

    predict_result = Result()
    predict_sid = worker.subscribe(predict_result.handle_event, tag="p1")
    worker.predict({}, tag="p1").result()

    assert predict_result.done
    assert predict_result.output == "output"
    assert predict_result.stdout_lines == ["running prediction\n"]

    worker.unsubscribe(predict_sid)


@uses_worker(
    "setup_async_with_sync_predict",
    max_concurrency=1,
    min_python=(3, 11),
    is_async=False,
    setup=False,
)
def test_setup_async_with_sync_predict_raises_error(worker: Worker):
    fut = worker.setup()
    result = Result()
    worker.subscribe(result.handle_event)

    with pytest.raises(FatalWorkerException):
        fut.result()
    assert result.done
    assert result.done.error
    assert (
        result.done.error_detail
        == "Invalid predictor: to use an async setup method you must use an async predict method"
    )


@uses_worker("simple", max_concurrency=5, setup=False)
def test_concurrency_with_sync_predictor_raises_error(worker):
    fut = worker.setup()
    result = Result()
    worker.subscribe(result.handle_event)

    with pytest.raises(FatalWorkerException):
        fut.result()
    assert result.done
    assert result.done.error
    assert (
        result.done.error_detail
        == "max_concurrency > 1 requires an async predict function, e.g. `async def predict()`"
    )


@uses_worker("stream_redirector_race_condition")
def test_stream_redirector_race_condition(worker):
    """
    StreamRedirector and _ChildWorker are using the same pipe to send data. When
    there are multiple threads trying to write to the same pipe, it can cause
    data corruption by race condition. The data corruption will cause pipe
    receiver to raise an exception due to unpickling error.
    """
    for _ in range(5):
        result = _process(worker, lambda: worker.predict({}))
        assert not result.done.error


@pytest.mark.timeout(HYPOTHESIS_TEST_TIMEOUT)
@pytest.mark.parametrize(
    "worker,payloads,expected_metrics", METRICS_FIXTURES, indirect=["worker"]
)
@settings(suppress_health_check=[HealthCheck.function_scoped_fixture])
@given(data=st.data())
def test_metrics(worker, payloads, expected_metrics, data):
    """
    We should get the metrics we expect from predictors that emit metrics.
    """
    payload = data.draw(st.fixed_dictionaries(payloads))
    tag = uuid.uuid4().hex

    result = _process(worker, lambda: worker.predict(payload, tag=tag), tag=tag)

    assert result.metrics == expected_metrics


@pytest.mark.timeout(HYPOTHESIS_TEST_TIMEOUT)
@pytest.mark.parametrize(
    "worker,payloads,output_generator", OUTPUT_FIXTURES, indirect=["worker"]
)
@settings(suppress_health_check=[HealthCheck.function_scoped_fixture])
@given(data=st.data())
def test_output(worker, payloads, output_generator, data):
    """
    We should get the outputs we expect from predictors that generate output.

    Note that most of the validation work here is actually done in _process.
    """
    payload = data.draw(st.fixed_dictionaries(payloads))
    expected_output = output_generator(payload)

    result = _process(worker, lambda: worker.predict(payload))

    assert result.output == expected_output


@pytest.mark.parametrize(
    "worker,expected_stdout,expected_stderr",
    SETUP_LOGS_FIXTURES,
    indirect=["worker"],
)
def test_setup_logging(worker: Worker, expected_stdout, expected_stderr):
    """
    We should get the logs we expect from predictors that generate logs during
    setup.
    """
    result = _process(worker, worker.setup)
    assert not result.done.error

    assert result.stdout == expected_stdout
    assert result.stderr == expected_stderr


@uses_worker_configs(
    [
        WorkerConfig("import_err", setup=False),
        WorkerConfig("import_err", setup=False, min_python=(3, 11), is_async=True),
    ]
)
def test_predictor_load_error_logging(worker: Worker):
    """
    This test ensures that we capture standard output that occurrs when the predictor
    errors when it is loaded. Before setup or predict are even run.
    """
    result = _process(worker, worker.setup, swallow_exceptions=True)

    assert result.done.error
    assert result.done.error_detail == "No module named 'missing_module'"

    assert result.stdout == "writing to stdout at import time\n"
    stderr_lines = result.stderr.splitlines(keepends=True)
    assert stderr_lines[0] == "writing to stderr at import time\n"

    assert "python/tests/server/fixtures/import_err.py" in stderr_lines[-3]
    assert "line 6" in stderr_lines[-3]
    assert "import missing_module" in stderr_lines[-2]
    assert stderr_lines[-1] == "ModuleNotFoundError: No module named 'missing_module'\n"


@pytest.mark.parametrize(
    "worker,expected_stdout,expected_stderr",
    PREDICT_LOGS_FIXTURES,
    indirect=["worker"],
)
def test_predict_logging(worker, expected_stdout, expected_stderr):
    """
    We should get the logs we expect from predictors that generate logs during
    predict.
    """
    result = _process(worker, lambda: worker.predict({}))

    assert result.stdout == expected_stdout
    assert result.stderr == expected_stderr


@uses_worker_configs(SLEEP_NO_SETUP_FIXTURES)
def test_cancel_is_safe(worker: Worker):
    """
    Calls to cancel at any time should not result in unexpected things
    happening or the cancelation of unexpected predictions.
    """

    tag = None
    if worker.uses_concurrency:
        tag = "p1"

    for _ in range(50):
        worker.cancel(tag)

    result = _process(worker, worker.setup)
    assert not result.done.error

    for _ in range(50):
        worker.cancel(tag)

    result1 = _process(
        worker,
        lambda: worker.predict({"sleep": 0.5}, tag),
        swallow_exceptions=True,
        tag=tag,
    )

    for _ in range(50):
        worker.cancel(tag)

    result2 = _process(
        worker,
        lambda: worker.predict({"sleep": 0.1}, tag),
        swallow_exceptions=True,
        tag=tag,
    )

    assert not result1.exception
    assert not result1.done.canceled
    assert not result2.exception
    assert not result2.done.canceled
    assert result2.output == "done in 0.1 seconds"


@uses_worker_configs(SLEEP_NO_SETUP_FIXTURES)
def test_cancel_idempotency(worker: Worker):
    """
    Multiple calls to cancel within the same prediction, while not necessary or
    recommended, should still only result in a single cancelled prediction, and
    should not affect subsequent predictions.
    """

    tag = None
    if worker.uses_concurrency:
        tag = "p1"

    result = _process(worker, worker.setup)
    assert not result.done.error

    fut = worker.predict({"sleep": 0.5}, tag)
    # We call cancel a WHOLE BUNCH to make sure that we don't propagate any
    # of those cancelations to subsequent predictions, regardless of the
    # internal implementation of exceptions raised inside signal handlers.
    for _ in range(5):
        time.sleep(0.05)
        for _ in range(100):
            worker.cancel(tag)
    result1 = fut.result()
    assert result1.canceled

    tag = None
    if worker.uses_concurrency:
        tag = "p2"
    result2 = _process(worker, lambda: worker.predict({"sleep": 0.1}, tag))

    assert not result2.done.canceled
    assert result2.output == "done in 0.1 seconds"


@uses_worker_configs(
    [
        WorkerConfig("sleep"),
        WorkerConfig("sleep_async", min_python=(3, 11), is_async=True),
        WorkerConfig(
            "sleep_async", min_python=(3, 11), is_async=True, max_concurrency=5
        ),
    ]
)
def test_cancel_multiple_predictions(worker: Worker):
    """
    Multiple predictions cancelled in a row shouldn't be a problem. This test
    is mainly ensuring that the _allow_cancel latch in Worker is correctly
    reset every time a prediction starts.
    """
    dones: list[Done] = []
    for i in range(5):
        tag = None
        if worker._max_concurrency > 1:
            tag = f"p{i}"
        fut = worker.predict({"sleep": 0.2}, tag)
        time.sleep(0.1)
        worker.cancel(tag)
        dones.append(fut.result())

    assert dones == [Done(canceled=True)] * 5

    assert not worker.predict({"sleep": 0}, "p6").result().canceled


@uses_worker_configs(
    [
        WorkerConfig(
            "sleep_async", min_python=(3, 11), is_async=True, max_concurrency=5
        ),
    ]
)
def test_cancel_some_predictions_async_with_concurrency(worker: Worker):
    """
    Multiple predictions cancelled in a row shouldn't be a problem. This test
    is mainly ensuring that the _allow_cancel latch in Worker is correctly
    reset every time a prediction starts.
    """
    fut1 = worker.predict({"sleep": 0.2}, "p1")
    fut2 = worker.predict({"sleep": 0.2}, "p2")
    fut3 = worker.predict({"sleep": 0.2}, "p3")

    time.sleep(0.1)

    worker.cancel("p2")

    assert not fut1.result().canceled
    assert fut2.result().canceled
    assert not fut3.result().canceled


@uses_worker_configs(SLEEP_FIXTURES)
def test_graceful_shutdown(worker: Worker):
    """
    On shutdown, the worker should finish running the current prediction, and
    then exit.
    """

    tag = None
    if worker.uses_concurrency:
        tag = "p1"

    saw_first_event = threading.Event()

    # When we see the first event, we'll start the shutdown process.
    worker.subscribe(lambda event: saw_first_event.set(), tag=tag)

    fut = worker.predict({"sleep": 1}, tag)

    saw_first_event.wait(timeout=1)
    worker.shutdown(timeout=2)

    assert fut.result() == Done()


@uses_worker("async_setup_uses_same_loop_as_predict", min_python=(3, 11), is_async=True)
def test_async_setup_uses_same_loop_as_predict(worker: Worker):
    result = _process(worker, lambda: worker.predict({}), tag=None)
    assert result, "Expected worker to return True to assert same event loop"


@frozen
class SetupState:
    fut: "Future[Done]"
    result: Result
    sid: int

    error: bool = False


@frozen
class PredictState:
    tag: Optional[str]
    payload: Dict[str, Any]
    fut: "Future[Done]"
    result: Result
    sid: int

    canceled: bool = False
    error: bool = False


class FakeChildWorker:
    exitcode = None
    alive = True
    pid: int = 0

    def start(self):
        pass

    def is_alive(self):
        return self.alive

    def send_cancel_signal(self):
        pass

    def terminate(self):
        pass

    def join(self):
        pass


class WorkerStateMachine(RuleBasedStateMachine):
    """
    This is a Hypothesis-driven rule-based state machine test. It is intended
    to ensure that any sequence of calls to the public API of Worker leaves the
    instance in an expected state.

    In short: any call should either throw InvalidStateException or should do
    what the caller asked.

    See https://hypothesis.readthedocs.io/en/latest/stateful.html for more on
    stateful testing with Hypothesis.
    """

    predict_pending = Bundle("predict_pending")
    predict_complete = Bundle("predict_complete")
    setup_pending = Bundle("setup_pending")
    setup_complete = Bundle("setup_complete")

    def __init__(self):
        super().__init__()

        parent_conn, child_conn = multiprocessing.get_context("spawn").Pipe()

        self.child = FakeChildWorker()
        self.child_events = child_conn

        self.pending = threading.Semaphore(0)

        self.worker = Worker(child=self.child, events=parent_conn, max_concurrency=4)

    def simulate_events(self, events, event_seen: threading.Event, *, tag=None):
        for event in events:
            event_seen.clear()
            self.child_events.send(Envelope(event, tag=tag))
            event_seen.wait(timeout=0.5)

    @rule(target=setup_pending)
    def setup(self):
        try:
            fut = self.worker.setup()
        except InvalidStateException:
            return multiple()
        else:
            result = Result()
            sid = self.worker.subscribe(result.handle_event)
            return SetupState(fut=fut, result=result, sid=sid)

    @rule(
        state=setup_pending,
        text=st.text(),
        source=st.sampled_from(["stdout", "stderr"]),
    )
    def simulate_setup_logs(self, state: SetupState, text: str, source: str):
        events = [Log(source=source, message=text)]
        self.simulate_events(events, event_seen=state.result.event_seen)

    @rule(state=consumes(setup_pending), target=setup_complete)
    def simulate_setup_success(self, state: SetupState):
        try:
            self.simulate_events(events=[Done()], event_seen=state.result.event_seen)
            return state
        finally:
            self.worker.unsubscribe(state.sid)

    @rule(state=consumes(setup_pending), target=setup_complete)
    def simulate_setup_failure(self, state: SetupState):
        try:
            self.simulate_events(
                events=[Done(error=True, error_detail="Setup failed!")],
                event_seen=state.result.event_seen,
            )
            return evolve(state, error=True)
        finally:
            self.worker.unsubscribe(state.sid)

    @rule(state=consumes(setup_complete))
    def await_setup(self, state: SetupState):
        if state.error:
            with pytest.raises(FatalWorkerException):
                state.fut.result()
            assert state.result.done.error
            assert state.result.done.error_detail == "Setup failed!"
        else:
            ev = state.fut.result()
            assert isinstance(ev, Done)
            assert state.result.done == Done()

    @rule(
        target=predict_pending,
        name=ST_NAMES,
        tag=st.uuids(),
        steps=st.integers(min_value=0, max_value=5),
    )
    def predict(self, name: str, steps: int, tag: uuid.UUID) -> PredictState:
        payload = {"name": name, "steps": steps}
        try:
            fut = self.worker.predict(payload, tag=tag.hex)
        except InvalidStateException:
            return multiple()
        else:
            # ensure the PredictionInput event has been sent; this needs to
            # happen before any further rules fire so we don't simulate a
            # prediction Done event before it has even started - that really
            # confuses the Worker
            assert self.child_events.poll(timeout=0.5)
            e = self.child_events.recv()
            assert isinstance(e, Envelope)
            assert isinstance(e.event, PredictionInput)
            assert e.tag == tag.hex
            assert not self.child_events.poll(timeout=0.1)

            result = Result()
            sid = self.worker.subscribe(result.handle_event, tag=tag.hex)
            return PredictState(
                tag=tag.hex, payload=payload, fut=fut, result=result, sid=sid
            )

    @rule(
        state=predict_pending,
        text=st.text(),
        source=st.sampled_from(["stdout", "stderr"]),
    )
    def simulate_predict_logs(self, state: PredictState, text: str, source: str):
        events = [Log(source=source, message=text)]
        self.simulate_events(events, event_seen=state.result.event_seen, tag=state.tag)

    @rule(state=consumes(predict_pending), target=predict_complete)
    def simulate_predict_success(self, state: PredictState):
        events = []

        steps = state.payload["steps"]
        name = state.payload["name"]

        if steps == 1:
            events.append(PredictionOutputType(multi=False))
            events.append(PredictionOutput(payload=f"NAME={name}"))

        elif steps > 1:
            events.append(PredictionOutputType(multi=True))
            for i in range(steps):
                events.append(
                    PredictionOutput(payload=f"NAME={name},STEP={i + 1}"),
                )

        events.append(Done(canceled=state.canceled))

        self.simulate_events(events, event_seen=state.result.event_seen, tag=state.tag)
        return state

    @rule(state=consumes(predict_pending), target=predict_complete)
    def simulate_predict_failure(self, state: PredictState):
        events = [
            Done(
                error=True,
                error_detail="Kaboom!",
                canceled=state.canceled,
            ),
        ]

        self.simulate_events(events, event_seen=state.result.event_seen, tag=state.tag)
        return evolve(state, error=True)

    @rule(state=consumes(predict_complete))
    def await_predict(self, state: PredictState):
        try:
            ev = state.fut.result()
            assert isinstance(ev, Done)
            assert state.result.done

            if state.canceled:
                assert state.result.done.canceled
                return

            if state.error:
                assert state.result.done.error
                assert state.result.done.error_detail == "Kaboom!"
                return

            steps = state.payload["steps"]
            name = state.payload["name"]

            if steps == 0:
                assert not state.result.output
            elif steps == 1:
                assert state.result.output == f"NAME={name}"
            else:
                assert state.result.output == [
                    f"NAME={name},STEP={i + 1}" for i in range(steps)
                ]

            assert state.result.done == Done()
        finally:
            self.worker.unsubscribe(state.sid)

    # For now, we only try canceling when we know a prediction is running.
    @rule(
        target=predict_pending,
        state=consumes(predict_pending),
    )
    def cancel(self, state: PredictState):
        self.worker.cancel(tag=state.tag)

        if not state.canceled:
            # if this prediction has not previously been canceled, Worker will
            # send a Cancel event to the child. We need to consume this event to
            # ensure we stay synced up on the child connection
            assert self.child_events.poll(timeout=0.5)
            e = self.child_events.recv()
            assert isinstance(e, Envelope)
            assert isinstance(e.event, Cancel)
            assert e.tag == state.tag

        return evolve(state, canceled=True)

    def teardown(self):
        self.child.alive = False
        self.worker.shutdown()


# Set a longer timeout for the state machine test. It can take a little while,
# particularly in CI, and particularly if it finds a failure, as shrinking
# might not happen all that quickly.
TestWorkerState = pytest.mark.timeout(600)(WorkerStateMachine.TestCase)



import asyncio

class Predictor:
    async def setup(self) -> None:
        self.loop = asyncio.get_running_loop()

    async def predict(self) -> bool:
        return self.loop == asyncio.get_running_loop()



import atexit
import multiprocessing
import pathlib
import signal
import subprocess
import sys
import time

from cog.types import Path
from cog import BasePredictor

from bg import ponger

def cleanup():
    for tmp in pathlib.Path("./").glob("*.tmp"):
        tmp.unlink(missing_ok=True)


atexit.register(cleanup)


class Predictor(BasePredictor):
    """
    This predictor checks the case where a process is spawned during setup via
    multiprocessing and then each prediction causes that process to write to stdout.
    """

    def setup(self) -> None:
        print("---> starting background process")

        cleanup()

        self.parent_conn, self.child_conn = multiprocessing.Pipe()
        self.lock = multiprocessing.Lock()
        self.bg = multiprocessing.Process(
            target=ponger, args=(self.child_conn, self.lock)
        )
        self.bg.start()

        print(f"---> started background process pid={self.bg.pid}")

    def predict(self, s: str) -> Path:
        if self.bg.is_alive():
            print(f"---> sending ping to background job pid={self.bg.pid}")

            self.child_conn.send("ping")

            print(f"---> sent ping to background job pid={self.bg.pid}")

            pong = self.parent_conn.recv()

            print(f"---> received {pong} from background job pid={self.bg.pid}")
        else:
            print(f"---> background job died status={status}")

            raise SystemExit

        out = Path(f"cog-test-integration-out.{time.time_ns()}.tmp")
        out.write_text("hello " + s)

        print(f"---> wrote output file {out}")

        return out



from cog import BasePredictor, Input
from typing import Literal


class Predictor(BasePredictor):
    def predict(self, text: Literal["foo", "bar"]) -> str:
        assert type(text) == str
        return text



from cog import BasePredictor, File


class Predictor(BasePredictor):
    def predict(self, files: list[File]) -> str:
        output_parts = []  # Use a list to collect file contents
        for f in files:
            # Assuming file content is in bytes, decode to str before appending
            content = f.read()
            if isinstance(content, bytes):
                # Decode bytes to str assuming UTF-8 encoding; adjust if needed
                content = content.decode('utf-8')
            output_parts.append(content)
        return "\n\n".join(output_parts)



import os
from functools import wraps
from typing import Any, Callable, Optional, TypeVar, Union, get_args, get_origin

R = TypeVar("R")


def env_property(
    env_var: str,
) -> Callable[[Callable[[Any], R]], Callable[[Any], R]]:
    """Wraps a class property in an environment variable check."""

    def decorator(func: Callable[[Any], R]) -> Callable[[Any], R]:
        @wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> R:
            result = os.environ.get(env_var)
            if result is not None:
                expected_type = func.__annotations__.get("return", str)
                if (
                    get_origin(expected_type) is Optional
                    or get_origin(expected_type) is Union
                ):
                    expected_type = get_args(expected_type)[0]
                return expected_type(result)
            result = func(*args, **kwargs)
            return result

        return wrapper

    return decorator



import base64
import io

import responses
from responses.matchers import multipart_matcher

from cog.types import PYDANTIC_V2

from .conftest import uses_predictor, uses_predictor_with_client_options


@uses_predictor("output_wrong_type")
def test_return_wrong_type(client):
    resp = client.post("/predictions")
    assert resp.status_code == 500


@uses_predictor("output_file")
def test_output_file(client, match):
    res = client.post("/predictions")
    assert res.status_code == 200
    assert res.json() == match(
        {
            "status": "succeeded",
            "output": "data:application/octet-stream;base64,aGVsbG8=",  # hello
        }
    )


@responses.activate
@uses_predictor("output_file_named")
def test_output_file_to_http(client, match):
    responses.add(
        responses.PUT,
        "http://example.com/upload/foo.txt",
        status=201,
        match=[multipart_matcher({"file": ("foo.txt", b"hello")})],
    )

    res = client.post(
        "/predictions", json={"output_file_prefix": "http://example.com/upload/"}
    )
    assert res.json() == match(
        {
            "status": "succeeded",
            "output": "http://example.com/upload/foo.txt",
        }
    )
    assert res.status_code == 200


@responses.activate
@uses_predictor_with_client_options("output_file_named", upload_url="https://dontuseme")
def test_output_file_to_http_with_upload_url_specified(client, match):
    # Ensure that even when --upload-url is provided on the command line,
    # uploads continue to go to the specified output_file_prefix, for backwards
    # compatibility.
    responses.add(
        responses.PUT,
        "http://example.com/upload/foo.txt",
        status=201,
        match=[multipart_matcher({"file": ("foo.txt", b"hello")})],
    )

    res = client.post(
        "/predictions", json={"output_file_prefix": "http://example.com/upload/"}
    )
    assert res.json() == match(
        {
            "status": "succeeded",
            "output": "http://example.com/upload/foo.txt",
        }
    )
    assert res.status_code == 200


@uses_predictor("output_path_image")
def test_output_path(client):
    res = client.post("/predictions")
    assert res.status_code == 200
    header, b64data = res.json()["output"].split(",", 1)
    # need both image/bmp and image/x-ms-bmp until https://bugs.python.org/issue44211 is fixed
    assert header in ["data:image/bmp;base64", "data:image/x-ms-bmp;base64"]
    assert len(base64.b64decode(b64data)) == 195894


@responses.activate
@uses_predictor("output_path_text")
def test_output_path_to_http(client, match):
    fh = io.BytesIO(b"hello")
    fh.name = "file.txt"
    responses.add(
        responses.PUT,
        "http://example.com/upload/file.txt",
        status=201,
        match=[multipart_matcher({"file": fh})],
    )

    res = client.post(
        "/predictions", json={"output_file_prefix": "http://example.com/upload/"}
    )
    assert res.json() == match(
        {
            "status": "succeeded",
            "output": "http://example.com/upload/file.txt",
        }
    )
    assert res.status_code == 200


@uses_predictor("output_complex")
def test_complex_output(client, match):
    resp = client.post("/predictions")
    assert resp.json() == match(
        {
            "output": {
                "file": "data:application/octet-stream;base64,aGVsbG8=",
                "text": "hello",
            },
            "status": "succeeded",
        }
    )
    assert resp.status_code == 200


@uses_predictor("output_iterator_complex")
def test_iterator_of_list_of_complex_output(client, match):
    resp = client.post("/predictions")
    assert resp.json() == match(
        {
            "output": [[{"text": "hello"}]],
            "status": "succeeded",
        }
    )
    assert resp.status_code == 200


if not PYDANTIC_V2:

    @uses_predictor("output_numpy")
    def test_json_output_numpy(client, match):
        resp = client.post("/predictions")
        assert resp.status_code == 200
        assert resp.json() == match({"output": 1.0, "status": "succeeded"})



from cog import BasePredictor, ConcatenateIterator


class Predictor(BasePredictor):
    def predict(
        self,
    ) -> ConcatenateIterator[str]:
        predictions = ["foo", "bar", "baz"]
        for prediction in predictions:
            yield prediction



import json
import os
import subprocess
from pathlib import Path

import pytest

from .util import assert_versions_match


def test_build_without_predictor(docker_image):
    project_dir = Path(__file__).parent / "fixtures/no-predictor-project"
    build_process = subprocess.run(
        ["cog", "build", "-t", docker_image],
        cwd=project_dir,
        capture_output=True,
    )
    assert build_process.returncode > 0
    assert (
        "Can\\'t run predictions: \\'predict\\' option not found"
        in build_process.stderr.decode()
    )


def test_build_names_uses_image_option_in_cog_yaml(tmpdir, docker_image):
    with open(tmpdir / "cog.yaml", "w") as f:
        cog_yaml = f"""
image: {docker_image}
build:
  python_version: 3.8
predict: predict.py:Predictor
"""
        f.write(cog_yaml)

    with open(tmpdir / "predict.py", "w") as f:
        code = """
from cog import BasePredictor

class Predictor(BasePredictor):
    def predict(self, text: str) -> str:
        return text

"""
        f.write(code)

    subprocess.run(
        ["cog", "build"],
        cwd=tmpdir,
        check=True,
    )
    assert docker_image in str(
        subprocess.run(["docker", "images"], capture_output=True, check=True).stdout
    )


def test_build_with_model(docker_image):
    project_dir = Path(__file__).parent / "fixtures/path-project"
    subprocess.run(
        ["cog", "build", "-t", docker_image],
        cwd=project_dir,
        check=True,
    )
    image = json.loads(
        subprocess.run(
            ["docker", "image", "inspect", docker_image],
            capture_output=True,
            check=True,
        ).stdout
    )
    labels = image[0]["Config"]["Labels"]
    schema = json.loads(labels["run.cog.openapi_schema"])

    assert schema["components"]["schemas"]["Input"] == {
        "title": "Input",
        "required": ["text", "path"],
        "type": "object",
        "properties": {
            "text": {"title": "Text", "type": "string", "x-order": 0},
            "path": {
                "title": "Path",
                "type": "string",
                "format": "uri",
                "x-order": 1,
            },
        },
    }


def test_build_invalid_schema(docker_image):
    project_dir = Path(__file__).parent / "fixtures/invalid-int-project"
    build_process = subprocess.run(
        ["cog", "build", "-t", docker_image],
        cwd=project_dir,
        capture_output=True,
    )
    assert build_process.returncode > 0
    assert "invalid default: number must be at least 2" in build_process.stderr.decode()


@pytest.mark.skipif(os.environ.get("CI") != "true", reason="only runs in CI")
def test_build_gpu_model_on_cpu(tmpdir, docker_image):
    with open(tmpdir / "cog.yaml", "w") as f:
        cog_yaml = """
build:
  python_version: 3.8
  gpu: true
predict: predict.py:Predictor
"""
        f.write(cog_yaml)

    with open(tmpdir / "predict.py", "w") as f:
        code = """
from cog import BasePredictor

class Predictor(BasePredictor):
    def predict(self, text: str) -> str:
        return text

"""
        f.write(code)

    subprocess.run(
        ["git", "config", "--global", "user.email", "noreply@replicate.com"],
        cwd=tmpdir,
        check=True,
    )

    subprocess.run(
        ["git", "config", "--global", "user.name", "Replicate Test Bot"],
        cwd=tmpdir,
        check=True,
    )

    subprocess.run(
        ["git", "init"],
        cwd=tmpdir,
        check=True,
    )
    subprocess.run(
        ["git", "commit", "--allow-empty", "-m", "initial"],
        cwd=tmpdir,
        check=True,
    )
    subprocess.run(
        ["git", "tag", "0.0.1"],
        cwd=tmpdir,
        check=True,
    )

    subprocess.run(
        ["cog", "build", "-t", docker_image],
        cwd=tmpdir,
        check=True,
    )
    assert docker_image in str(
        subprocess.run(["docker", "images"], capture_output=True, check=True).stdout
    )
    image = json.loads(
        subprocess.run(
            ["docker", "image", "inspect", docker_image],
            capture_output=True,
            check=True,
        ).stdout
    )
    labels = image[0]["Config"]["Labels"]

    assert len(labels["run.cog.version"]) > 0
    assert json.loads(labels["run.cog.config"]) == {
        "build": {
            "python_version": "3.8",
            "gpu": True,
            "cuda": "11.8",
            "cudnn": "8",
        },
        "predict": "predict.py:Predictor",
    }
    assert "run.cog.openapi_schema" in labels

    assert len(labels["org.opencontainers.image.version"]) > 0
    assert len(labels["org.opencontainers.image.revision"]) > 0


def test_build_with_cog_init_templates(tmpdir, docker_image):
    subprocess.run(
        ["cog", "init"],
        cwd=tmpdir,
        capture_output=True,
        check=True,
    )

    build_process = subprocess.run(
        ["cog", "build", "-t", docker_image],
        cwd=tmpdir,
        capture_output=True,
        check=True,
    )

    assert build_process.returncode == 0
    assert "Image built as cog-" in build_process.stderr.decode()


def test_build_with_complex_output(tmpdir, docker_image):
    project_dir = Path(__file__).parent / "fixtures/complex_output_project"
    build_process = subprocess.run(
        ["cog", "build", "-t", docker_image],
        cwd=project_dir,
        capture_output=True,
    )
    assert build_process.returncode == 0
    assert "Image built as cog-" in build_process.stderr.decode()


def test_python_37_deprecated(docker_image):
    project_dir = Path(__file__).parent / "fixtures/python_37"
    build_process = subprocess.run(
        ["cog", "build", "-t", docker_image],
        cwd=project_dir,
        capture_output=True,
    )
    assert build_process.returncode > 0
    assert (
        "minimum supported Python version is 3.8. requested 3.7"
        in build_process.stderr.decode()
    )


def test_build_base_image_sha(docker_image):
    project_dir = Path(__file__).parent / "fixtures/path-project"
    subprocess.run(
        ["cog", "build", "-t", docker_image, "--use-cog-base-image"],
        cwd=project_dir,
        check=True,
    )
    image = json.loads(
        subprocess.run(
            ["docker", "image", "inspect", docker_image],
            capture_output=True,
            check=True,
        ).stdout
    )
    labels = image[0]["Config"]["Labels"]
    base_layer_hash = labels["run.cog.cog-base-image-last-layer-sha"]
    layers = image[0]["RootFS"]["Layers"]
    assert base_layer_hash in layers


def test_torch_2_0_3_cu118_base_image(docker_image):
    project_dir = Path(__file__).parent / "fixtures/torch-cuda-baseimage-project"
    build_process = subprocess.run(
        ["cog", "build", "-t", docker_image, "--use-cog-base-image"],
        cwd=project_dir,
        capture_output=True,
    )
    assert build_process.returncode == 0


def test_torch_1_13_0_base_image_fallback(docker_image):
    project_dir = Path(__file__).parent / "fixtures/torch-baseimage-project"
    build_process = subprocess.run(
        ["cog", "build", "-t", docker_image, "--openapi-schema", "openapi.json"],
        cwd=project_dir,
        capture_output=True,
    )
    assert build_process.returncode == 0


def test_torch_1_13_0_base_image_fail(docker_image):
    project_dir = Path(__file__).parent / "fixtures/torch-baseimage-project"
    build_process = subprocess.run(
        [
            "cog",
            "build",
            "-t",
            docker_image,
            "--openapi-schema",
            "openapi.json",
            "--use-cog-base-image",
        ],
        cwd=project_dir,
        capture_output=True,
    )
    assert build_process.returncode == 1


def test_torch_1_13_0_base_image_fail_explicit(docker_image):
    project_dir = Path(__file__).parent / "fixtures/torch-baseimage-project"
    build_process = subprocess.run(
        [
            "cog",
            "build",
            "-t",
            docker_image,
            "--openapi-schema",
            "openapi.json",
            "--use-cog-base-image=false",
        ],
        cwd=project_dir,
        capture_output=True,
    )
    assert build_process.returncode == 0


def test_precompile(docker_image):
    project_dir = Path(__file__).parent / "fixtures/torch-baseimage-project"
    build_process = subprocess.run(
        [
            "cog",
            "build",
            "-t",
            docker_image,
            "--openapi-schema",
            "openapi.json",
            "--use-cog-base-image=false",
            "--precompile",
        ],
        cwd=project_dir,
        capture_output=True,
    )
    assert build_process.returncode == 0


def test_cog_install_base_image(docker_image):
    project_dir = Path(__file__).parent / "fixtures/string-project"
    build_process = subprocess.run(
        [
            "cog",
            "build",
            "-t",
            docker_image,
            "--use-cog-base-image=true",
        ],
        cwd=project_dir,
        capture_output=True,
    )
    assert build_process.returncode == 0
    cog_installed_version_process = subprocess.run(
        [
            "docker",
            "run",
            "-t",
            docker_image,
            "python",
            "-c",
            "import cog; print(cog.__version__)",
        ],
        cwd=project_dir,
        capture_output=True,
    )
    assert cog_installed_version_process.returncode == 0
    cog_installed_version = cog_installed_version_process.stdout.decode().strip()
    cog_version_process = subprocess.run(
        [
            "cog",
            "--version",
        ],
        cwd=project_dir,
        capture_output=True,
    )
    cog_version = cog_version_process.stdout.decode().strip().split()[2]

    assert_versions_match(
        semver_version=cog_version,
        pep440_version=cog_installed_version,
    )


def test_pip_freeze(docker_image):
    project_dir = Path(__file__).parent / "fixtures/path-project"
    subprocess.run(
        ["cog", "build", "-t", docker_image],
        cwd=project_dir,
        check=True,
    )
    image = json.loads(
        subprocess.run(
            ["docker", "image", "inspect", docker_image],
            capture_output=True,
            check=True,
        ).stdout
    )
    labels = image[0]["Config"]["Labels"]
    pip_freeze = labels["run.cog.pip_freeze"]
    pip_freeze = "\n".join(
        [
            x
            for x in pip_freeze.split("\n")
            if not x.startswith("cog @")
            and not x.startswith("fastapi")
            and not x.startswith("starlette")
        ]
    )
    assert (
        pip_freeze
        == "anyio==4.4.0\nattrs==23.2.0\ncertifi==2024.8.30\ncharset-normalizer==3.3.2\nclick==8.1.7\nexceptiongroup==1.2.2\nh11==0.14.0\nhttptools==0.6.1\nidna==3.8\npydantic==1.10.18\npython-dotenv==1.0.1\nPyYAML==6.0.2\nrequests==2.32.3\nsniffio==1.3.1\nstructlog==24.4.0\ntyping_extensions==4.12.2\nurllib3==2.2.2\nuvicorn==0.30.6\nuvloop==0.20.0\nwatchfiles==0.24.0\nwebsockets==13.0.1\n"
    )


def test_cog_installs_apt_packages(docker_image):
    project_dir = Path(__file__).parent / "fixtures/apt-packages"
    build_process = subprocess.run(
        [
            "cog",
            "build",
            "-t",
            docker_image,
        ],
        cwd=project_dir,
        capture_output=True,
    )
    # Test that the build completes successfully.
    # If the apt-packages weren't installed the run command would fail.
    assert build_process.returncode == 0


def test_fast_build(docker_image):
    project_dir = Path(__file__).parent / "fixtures/fast-build"
    weights_file = os.path.join(project_dir, "weights.h5")
    with open(weights_file, "w", encoding="utf8") as handle:
        handle.seek(256 * 1024 * 1024)
        handle.write("\0")

    build_process = subprocess.run(
        ["cog", "build", "-t", docker_image, "--x-fast"],
        cwd=project_dir,
        capture_output=True,
    )

    os.remove(weights_file)

    assert build_process.returncode == 0


def test_pydantic2(docker_image):
    project_dir = Path(__file__).parent / "fixtures/pydantic2"

    build_process = subprocess.run(
        ["cog", "build", "-t", docker_image],
        cwd=project_dir,
        capture_output=True,
    )

    assert build_process.returncode == 0


def test_ffmpeg_base_image(docker_image):
    project_dir = Path(__file__).parent / "fixtures/ffmpeg-package"

    build_process = subprocess.run(
        ["cog", "build", "-t", docker_image],
        cwd=project_dir,
        capture_output=True,
    )

    assert build_process.returncode == 0


def test_bad_dockerignore(docker_image):
    project_dir = Path(__file__).parent / "fixtures/bad-dockerignore"

    build_process = subprocess.run(
        ["cog", "build", "-t", docker_image],
        cwd=project_dir,
        capture_output=True,
    )

    assert build_process.returncode == 1
    assert (
        "The .cog tmp path cannot be ignored by docker in .dockerignore"
        in build_process.stderr.decode()
    )


def test_pydantic1_none(docker_image):
    project_dir = Path(__file__).parent / "fixtures/pydantic1-none"

    build_process = subprocess.run(
        ["cog", "build", "-t", docker_image],
        cwd=project_dir,
        capture_output=True,
    )

    assert build_process.returncode == 0


def test_fast_build_with_local_image(docker_image):
    project_dir = Path(__file__).parent / "fixtures/fast-build"
    weights_file = os.path.join(project_dir, "weights.h5")
    with open(weights_file, "w", encoding="utf8") as handle:
        handle.seek(256 * 1024 * 1024)
        handle.write("\0")

    build_process = subprocess.run(
        ["cog", "build", "-t", docker_image, "--x-fast", "--x-localimage"],
        cwd=project_dir,
        capture_output=True,
    )

    os.remove(weights_file)

    assert build_process.returncode == 0


def test_local_whl_install(docker_image):
    project_dir = Path(__file__).parent / "fixtures/local-whl-install"

    build_process = subprocess.run(
        ["cog", "build", "-t", docker_image],
        cwd=project_dir,
        capture_output=True,
    )

    assert build_process.returncode == 0


def test_overrides(docker_image):
    project_dir = Path(__file__).parent / "fixtures/overrides-project"

    build_process = subprocess.run(
        ["cog", "build", "-t", docker_image],
        cwd=project_dir,
        capture_output=True,
    )

    assert build_process.returncode == 0



from __future__ import annotations

import contextlib
import ctypes
import io
import os
import re
import selectors
import sys
import threading
import uuid
from types import TracebackType
from typing import Any, BinaryIO, Callable, Dict, List, Sequence, TextIO, Union

import pydantic
from typing_extensions import Self  # added to typing in python 3.11

from ..types import PYDANTIC_V2
from .errors import CogRuntimeError, CogTimeoutError


class _SimpleStreamWrapper(io.TextIOWrapper):
    """
    _SimpleStreamWrapper wraps a binary I/O buffer and provides a TextIOWrapper
    interface (primarily write and flush methods) which call a provided
    callback function instead of (or, if `tee` is True, in addition to) writing
    to the underlying buffer.
    """

    def __init__(
        self,
        buffer: BinaryIO,
        callback: Callable[[str, str], None],
        tee: bool = False,
    ) -> None:
        super().__init__(buffer)

        self._callback = callback
        self._tee = tee
        self._buffer = []

    def write(self, s: str) -> int:
        length = len(s)
        self._buffer.append(s)
        if self._tee:
            super().write(s)

        if "\n" in s or "\r" in s:
            self.flush()

        return length

    def flush(self) -> None:
        self._callback(self.name, "".join(self._buffer))
        self._buffer.clear()
        if self._tee:
            super().flush()


class _StreamWrapper:
    def __init__(self, name: str, stream: TextIO) -> None:
        self.name = name
        self._stream = stream
        self._original_fp: TextIO | None = None
        self._wrapped_fp: TextIO | None = None

    def wrap(self) -> None:
        if self._wrapped_fp or self._original_fp:
            raise CogRuntimeError("stream is already wrapped")

        r, w = os.pipe()

        # Save a copy of the stream file descriptor.
        original_fd = self._stream.fileno()
        original_fd_copy = os.dup(original_fd)

        # Overwrite the stream file descriptor with the write end of the pipe.
        os.dup2(w, self._stream.fileno())
        os.close(w)

        # Create a writeable file object with the original FD. This can be used
        # to write to the original destination of the passed stream.
        self._original_fp = os.fdopen(original_fd_copy, "w")

        # Create a readable file object with the read end of the pipe. This can
        # be used to read any writes to the passed stream.
        #
        # We set the FD to be non-blocking so that we can select/poll/epoll
        # over multiple wrapped streams.
        os.set_blocking(r, False)
        self._wrapped_fp = os.fdopen(r, "r")

    def unwrap(self) -> None:
        if not self._wrapped_fp or not self._original_fp:
            raise CogRuntimeError("stream is not wrapped (call wrap first)")

        # Put the original file descriptor back.
        os.dup2(self._original_fp.fileno(), self._stream.fileno())

        # Close the write end of the pipe.
        self._original_fp.close()
        self._original_fp = None

        # Close the read end of the pipe.
        self._wrapped_fp.close()
        self._wrapped_fp = None

    def write(self, data: str) -> int:
        return self._stream.write(data)

    def flush(self) -> None:
        return self._stream.flush()

    @property
    def wrapped(self) -> TextIO:
        if not self._wrapped_fp:
            raise CogRuntimeError("stream is not wrapped (call wrap first)")
        return self._wrapped_fp

    @property
    def original(self) -> TextIO:
        if not self._original_fp:
            raise CogRuntimeError("stream is not wrapped (call wrap first)")
        return self._original_fp


if sys.version_info < (3, 9):

    class _SimpleStreamRedirectorBase(contextlib.AbstractContextManager):
        pass
else:

    class _SimpleStreamRedirectorBase(
        contextlib.AbstractContextManager["SimpleStreamRedirector"]
    ):
        pass


class SimpleStreamRedirector(_SimpleStreamRedirectorBase):
    """
    SimpleStreamRedirector is a context manager that redirects I/O streams to a
    callback function. If `tee` is True, it also writes output to the original
    streams.

    Unlike StreamRedirector, the underlying stream file descriptors are not
    modified, which means that only stream writes from Python code will be
    captured. Writes from native code will not be captured.

    Unlike StreamRedirector, the streams redirected cannot be configured. The
    context manager is only able to redirect STDOUT and STDERR.
    """

    def __init__(
        self,
        callback: Callable[[str, str], None],
        tee: bool = False,
    ) -> None:
        self._callback = callback
        self._tee = tee

        stdout_wrapper = _SimpleStreamWrapper(sys.stdout.buffer, callback, tee)
        stderr_wrapper = _SimpleStreamWrapper(sys.stderr.buffer, callback, tee)
        self._stdout_ctx = contextlib.redirect_stdout(stdout_wrapper)
        self._stderr_ctx = contextlib.redirect_stderr(stderr_wrapper)

    def __enter__(self) -> Self:
        self._stdout_ctx.__enter__()
        self._stderr_ctx.__enter__()
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        self._stdout_ctx.__exit__(exc_type, exc_value, traceback)
        self._stderr_ctx.__exit__(exc_type, exc_value, traceback)

    def drain(self, timeout: float = 0.0) -> None:
        # Draining isn't complicated for SimpleStreamRedirector, since we're not
        # moving data between threads. We just need to flush the streams.
        sys.stdout.flush()
        sys.stderr.flush()


if sys.version_info < (3, 9):

    class _StreamRedirectorBase(contextlib.AbstractContextManager):
        pass
else:

    class _StreamRedirectorBase(contextlib.AbstractContextManager["StreamRedirector"]):
        pass


class StreamRedirector(_StreamRedirectorBase):
    """
    StreamRedirector is a context manager that redirects I/O streams to a
    callback function. If `tee` is True, it also writes output to the original
    streams.

    If `streams` is not provided, it defaults to redirecting the process's
    STDOUT and STDERR file descriptors.
    """

    def __init__(
        self,
        callback: Callable[[str, str], None],
        tee: bool = False,
        streams: Sequence[TextIO] = None,
    ) -> None:
        self._callback = callback
        self._tee = tee

        self._depth = 0
        self._drain_token = uuid.uuid4().hex
        self._drain_event = threading.Event()
        self._terminate_token = uuid.uuid4().hex

        if not streams:
            streams = [sys.stdout, sys.stderr]
        self._streams = [_StreamWrapper(s.name, s) for s in streams]

    def __enter__(self) -> Self:
        self._depth += 1

        if self._depth == 1:
            for s in self._streams:
                s.wrap()

            self._thread = threading.Thread(target=self._start)
            self._thread.start()

        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        self._depth -= 1

        if self._depth == 0:
            self._stop()
            self._thread.join()

            for s in self._streams:
                s.unwrap()

    def drain(self, timeout: float = 1) -> None:
        self._drain_event.clear()
        for stream in self._streams:
            stream.write(self._drain_token + "\n")
            stream.flush()
        if not self._drain_event.wait(timeout=timeout):
            raise CogTimeoutError("output streams failed to drain")

    def _start(self) -> None:
        selector = selectors.DefaultSelector()

        should_exit = False
        drain_tokens_seen = 0
        drain_tokens_needed = 0
        buffers = {}

        for stream in self._streams:
            selector.register(stream.wrapped, selectors.EVENT_READ, stream)
            buffers[stream.name] = io.StringIO()
            drain_tokens_needed += 1

        while not should_exit:
            for key, _ in selector.select():
                stream = key.data

                for line in stream.wrapped:
                    if not line.endswith("\n"):
                        # TODO: limit how much we're prepared to buffer on a
                        # single line
                        buffers[stream.name].write(line)
                        continue

                    full_line = buffers[stream.name].getvalue() + line.strip()

                    # Reset buffer (this is quicker and easier than resetting
                    # the existing buffer, but may generate more garbage).
                    buffers[stream.name] = io.StringIO()

                    if full_line.endswith(self._terminate_token):
                        should_exit = True
                        full_line = full_line[: -len(self._terminate_token)]

                    if full_line.endswith(self._drain_token):
                        drain_tokens_seen += 1
                        full_line = full_line[: -len(self._drain_token)]

                    # If full_line is empty at this point it means the only
                    # thing in the line was a drain token (or a terminate
                    # token).
                    if full_line:
                        self._callback(stream.name, full_line + "\n")
                        if self._tee:
                            stream.original.write(full_line + "\n")
                            stream.original.flush()

                    if drain_tokens_seen >= drain_tokens_needed:
                        self._drain_event.set()
                        drain_tokens_seen = 0

    def _stop(self) -> None:
        for s in self._streams:
            s.write(self._terminate_token + "\n")
            s.flush()
            break  # we only need to send the terminate token to one stream


# Precompile the regular expression
_ADDRESS_PATTERN = re.compile(r"0x[0-9A-Fa-f]+")


if PYDANTIC_V2:

    def _unwrap_pydantic_serialization_iterator(obj: Any) -> Any:
        # SerializationIterator doesn't expose the object it wraps
        # but does give us a pointer in the `__repr__` string
        match = _ADDRESS_PATTERN.search(repr(obj))
        if match:
            address = int(match.group(), 16)
            # Cast the memory address to a Python object
            return ctypes.cast(address, ctypes.py_object).value

        return obj

    def unwrap_pydantic_serialization_iterators(obj: Any) -> Any:
        """
        Unwraps instances of `pydantic_core._pydantic_core.SerializationIterator`,
        returning their underlying object so that it can be pickled when passed
        between multiprocessing workers.

        This is a temporary workaround until the following issues are fixed:
        - https://github.com/pydantic/pydantic/issues/8907
        - https://github.com/pydantic/pydantic-core/pull/1399
        - https://github.com/pydantic/pydantic-core/pull/1401
        """

        if type(obj).__name__ == "SerializationIterator":
            return _unwrap_pydantic_serialization_iterator(obj)
        if type(obj) == str:  # noqa: E721 # pylint: disable=unidiomatic-typecheck
            return obj
        if isinstance(obj, pydantic.BaseModel):
            return unwrap_pydantic_serialization_iterators(
                obj.model_dump(exclude_unset=True)
            )
        if isinstance(obj, dict):
            return {
                key: unwrap_pydantic_serialization_iterators(value)
                for key, value in obj.items()
            }
        if isinstance(obj, list):
            return [unwrap_pydantic_serialization_iterators(value) for value in obj]
        return obj


def update_openapi_schema_for_pydantic_2(
    openapi_schema: Dict[str, Any],
) -> None:
    _remove_webhook_events_filter_title(openapi_schema)
    _remove_empty_or_nullable_anyof(openapi_schema)
    _flatten_selected_allof_refs(openapi_schema)
    _extract_enum_properties(openapi_schema)
    _set_default_enumeration_description(openapi_schema)
    _restore_allof_for_prediction_id_put(openapi_schema)


def _remove_webhook_events_filter_title(
    openapi_schema: Dict[str, Any],
) -> None:
    try:
        del openapi_schema["components"]["schemas"]["PredictionRequest"]["properties"][
            "webhook_events_filter"
        ]["title"]
    except KeyError:
        pass


def _remove_empty_or_nullable_anyof(
    openapi_schema: Union[Dict[str, Any], List[Dict[str, Any]]],
) -> None:
    if isinstance(openapi_schema, dict):
        for key, value in list(openapi_schema.items()):
            if key == "anyOf" and isinstance(value, list):
                non_null_types = [item for item in value if item.get("type") != "null"]
                if len(non_null_types) == 0:
                    del openapi_schema[key]
                elif len(non_null_types) == 1:
                    openapi_schema.update(non_null_types[0])
                    del openapi_schema[key]

                    # FIXME: Update tests to expect nullable
                    # openapi_schema["nullable"] = True

            else:
                _remove_empty_or_nullable_anyof(value)
    elif isinstance(openapi_schema, list):  # pyright: ignore
        for item in openapi_schema:
            _remove_empty_or_nullable_anyof(item)


def _flatten_selected_allof_refs(
    openapi_schema: Dict[str, Any],
) -> None:
    try:
        response = openapi_schema["components"]["schemas"]["PredictionResponse"]
        response["properties"]["output"] = {"$ref": "#/components/schemas/Output"}
    except KeyError:
        pass

    try:
        path = openapi_schema["paths"]["/predictions"]["post"]
        body = path["requestBody"]
        body["content"]["application/json"]["schema"] = {
            "$ref": "#/components/schemas/PredictionRequest"
        }
    except KeyError:
        pass


def _extract_enum_properties(
    openapi_schema: Dict[str, Any],
) -> None:
    schemas = openapi_schema.get("components", {}).get("schemas", {})
    if "Input" in schemas and "properties" in schemas["Input"]:
        input_properties = schemas["Input"]["properties"]
        for prop_name, prop_value in input_properties.items():
            if "enum" in prop_value:
                # Create a new schema for the enum
                schemas[prop_name] = {
                    "type": prop_value["type"],
                    "enum": prop_value["enum"],
                    "title": prop_name,
                    "description": prop_value.get("description", "An enumeration."),
                }

                # Replace the original property with an allOf reference
                input_properties[prop_name] = {
                    "allOf": [{"$ref": f"#/components/schemas/{prop_name}"}]
                }

                # Preserve x-order if it exists
                if "x-order" in prop_value:
                    input_properties[prop_name]["x-order"] = prop_value["x-order"]


def _set_default_enumeration_description(
    openapi_schema: Union[Dict[str, Any], List[Dict[str, Any]]],
) -> None:
    if isinstance(openapi_schema, dict):
        schemas = openapi_schema.get("components", {}).get("schemas", {})
        for _key, value in list(schemas.items()):
            if isinstance(value, dict) and value.get("enum"):
                value["description"] = value.get("description", "An enumeration.")
            else:
                _set_default_enumeration_description(value)
    elif isinstance(openapi_schema, list):  # pyright: ignore
        for item in openapi_schema:
            _set_default_enumeration_description(item)


def _restore_allof_for_prediction_id_put(
    openapi_schema: Dict[str, Any],
) -> None:
    try:
        put_operation = openapi_schema["paths"]["/predictions/{prediction_id}"]["put"]
        request_body = put_operation["requestBody"]
        json_schema = request_body["content"]["application/json"]["schema"]

        if "$ref" in json_schema:
            ref = json_schema["$ref"]
            json_schema.clear()
            json_schema["allOf"] = [{"$ref": ref}]
            json_schema["title"] = "Prediction Request"
    except KeyError:
        pass

    for _key, value in (
        openapi_schema.get("components", {})
        .get("schemas", {})
        .get("Input", {})
        .get("properties", {})
        .items()
    ):
        if "$ref" in value:
            ref = value["$ref"]
            del value["$ref"]
            value["allOf"] = [{"$ref": ref}]



from pydantic import BaseModel


class Output(BaseModel):
    number: int
    text: str


class Predictor:
    def setup(self):
        pass

    def predict(self) -> Output:
        return Output(number=42, text="meaning of life")



from typing import Any, Dict, Optional, Union

from attrs import define, field, validators


# From worker parent process
#
@define
class Cancel:
    pass


@define
class PredictionInput:
    payload: Dict[str, Any]


@define
class Shutdown:
    pass


# From predictor child process
#
@define
class Log:
    message: str
    source: str = field(validator=validators.in_(["stdout", "stderr"]))


@define
class PredictionMetric:
    name: str
    value: Union[float, int]


@define
class PredictionOutput:
    payload: Any


@define
class PredictionOutputType:
    multi: bool = False


@define
class Done:
    canceled: bool = False
    error: bool = False
    error_detail: str = ""


@define
class Envelope:
    """
    Envelope contains an arbitrary event along with an optional tag used to
    tangle/untangle concurrent work.
    """

    event: Union[
        Cancel,
        PredictionInput,
        Shutdown,
        Log,
        PredictionMetric,
        PredictionOutput,
        PredictionOutputType,
        Done,
    ]
    tag: Optional[str] = None



import functools

import pytest


class ObjectMatcher:
    def __init__(self, name, pattern):
        self.name = name
        self.pattern = pattern

    def __eq__(self, other):
        if not isinstance(other, dict):
            return self.pattern == other
        minimal = {k: other[k] for k in self.pattern.keys() if k in other}
        return self.pattern == minimal

    def __repr__(self):
        return f"{self.name}({repr(self.pattern)})"


@pytest.fixture
def match():
    return functools.partial(ObjectMatcher, "match")



from cog import BasePredictor, Input
from typing import Optional


class Predictor(BasePredictor):
    def predict(self, s: Optional[str] = Input(description="Hello String.")) -> str:
        return "hello " + (s if s is not None else "No One")



from typing import Iterator

from cog import BasePredictor


class Predictor(BasePredictor):
    def predict(self) -> Iterator[str]:
        predictions = ["foo", "bar", "baz"]
        for prediction in predictions:
            yield prediction



import os
import sys
import tempfile
import threading
import time
from pathlib import Path

from cog.wait import (
    COG_EAGER_IMPORTS_ENV_VAR,
    COG_PYENV_PATH_ENV_VAR,
    COG_WAIT_FILE_ENV_VAR,
    PYTHON_VERSION_ENV_VAR,
    PYTHONPATH_ENV_VAR,
    eagerly_import_modules,
    wait_for_env,
    wait_for_file,
)


def test_wait_for_file_no_env_var():
    if COG_WAIT_FILE_ENV_VAR in os.environ:
        del os.environ[COG_WAIT_FILE_ENV_VAR]
    result = wait_for_file()
    assert result, "We should immediately return when no wait file is specified."


def test_wait_for_file_exists():
    with tempfile.NamedTemporaryFile() as tmpfile:
        os.environ[COG_WAIT_FILE_ENV_VAR] = tmpfile.name
        result = wait_for_file(timeout=5.0)
        del os.environ[COG_WAIT_FILE_ENV_VAR]
        assert result, "We should immediately return when the file already exists."


def test_wait_for_file_waits_for_file():
    wait_file = os.path.join(os.path.dirname(__file__), "flag_file")
    if os.path.exists(wait_file):
        os.remove(wait_file)
    os.environ[COG_WAIT_FILE_ENV_VAR] = wait_file

    def create_file():
        time.sleep(2.0)
        Path(wait_file).touch()

    thread = threading.Thread(target=create_file)
    thread.start()
    result = wait_for_file(timeout=5.0)
    del os.environ[COG_WAIT_FILE_ENV_VAR]
    os.remove(wait_file)
    assert result, "We should return when the file is touched."


def test_wait_for_file_timeout():
    os.environ[COG_WAIT_FILE_ENV_VAR] = os.path.join(
        os.path.dirname(__file__), "a_file_unknown"
    )
    result = wait_for_file(timeout=5.0)
    del os.environ[COG_WAIT_FILE_ENV_VAR]
    assert not result, "We should return false when the timeout triggers."


def test_eagerly_import_modules_no_env_var():
    if COG_EAGER_IMPORTS_ENV_VAR in os.environ:
        del os.environ[COG_EAGER_IMPORTS_ENV_VAR]
    eagerly_import_modules()


def test_eagerly_import_modules():
    os.environ[COG_EAGER_IMPORTS_ENV_VAR] = "pytest,pathlib,time"
    import_count = eagerly_import_modules()
    del os.environ[COG_EAGER_IMPORTS_ENV_VAR]
    assert import_count == 3, "There should be 3 imports performed"


def test_wait_for_env_no_env_vars():
    if COG_WAIT_FILE_ENV_VAR in os.environ:
        del os.environ[COG_WAIT_FILE_ENV_VAR]
    if COG_EAGER_IMPORTS_ENV_VAR in os.environ:
        del os.environ[COG_EAGER_IMPORTS_ENV_VAR]
    result = wait_for_env()
    assert result, (
        "We should return true if we have no env vars associated with the wait."
    )


def test_wait_for_env():
    with tempfile.NamedTemporaryFile() as tmpfile:
        os.environ[COG_WAIT_FILE_ENV_VAR] = tmpfile.name
        os.environ[COG_EAGER_IMPORTS_ENV_VAR] = "pytest,pathlib,time"
        result = wait_for_env()
        assert result, (
            "We should return true if we have waited for the right environment."
        )
        del os.environ[COG_EAGER_IMPORTS_ENV_VAR]
        del os.environ[COG_WAIT_FILE_ENV_VAR]


def test_wait_inserts_pythonpath():
    with tempfile.NamedTemporaryFile() as tmpfile:
        original_sys_path = sys.path.copy()
        original_python_path = os.environ.get(PYTHONPATH_ENV_VAR)
        pyenv_path = os.path.dirname(tmpfile.name)
        os.environ[COG_WAIT_FILE_ENV_VAR] = tmpfile.name
        os.environ[COG_EAGER_IMPORTS_ENV_VAR] = "pytest,pathlib,time"
        os.environ[COG_PYENV_PATH_ENV_VAR] = pyenv_path
        os.environ[PYTHON_VERSION_ENV_VAR] = "3.11"
        wait_for_env()
        del os.environ[PYTHON_VERSION_ENV_VAR]
        del os.environ[COG_PYENV_PATH_ENV_VAR]
        del os.environ[COG_EAGER_IMPORTS_ENV_VAR]
        del os.environ[COG_WAIT_FILE_ENV_VAR]
        current_python_path = os.environ[PYTHONPATH_ENV_VAR]
        if original_python_path is None:
            del os.environ[PYTHONPATH_ENV_VAR]
        else:
            os.environ[PYTHONPATH_ENV_VAR] = original_python_path
        sys.path = original_sys_path
        expected_path = ":".join(
            original_sys_path + [pyenv_path + "/lib/python3.11/site-packages"]
        )
        assert expected_path == current_python_path, (
            "Our python path should be updated with the pyenv path."
        )



import logging
import os

import structlog
from structlog.typing import EventDict


def replace_level_with_severity(
    _: logging.Logger, __: str, event_dict: EventDict
) -> EventDict:
    """
    Replace the level field with a severity field as understood by Stackdriver
    logs.
    """
    if "level" in event_dict:
        event_dict["severity"] = event_dict.pop("level").upper()
    return event_dict


def setup_logging(*, log_level: int = logging.NOTSET) -> None:
    """
    Configure stdlib logger to use structlog processors and formatters so that
    uvicorn and application logs are consistent.
    """

    # Switch to human-friendly log output if LOG_FORMAT environment variable is
    # set to "development".
    development_logs = os.environ.get("LOG_FORMAT", "") == "development"

    processors: list[structlog.types.Processor] = [
        structlog.contextvars.merge_contextvars,
        structlog.stdlib.add_logger_name,
        structlog.stdlib.add_log_level,
        structlog.processors.StackInfoRenderer(),
        structlog.processors.TimeStamper(fmt="iso"),
    ]

    if development_logs:
        # In development, set `exc_info` on the log event if the log method is
        # `exception` and `exc_info` is not already set.
        #
        # Rendering of `exc_info` is handled by ConsoleRenderer.
        processors.append(structlog.dev.set_exc_info)
    else:
        # Outside of development mode `exc_info` must be set explicitly when
        # needed, and is translated into a formatted `exception` field.
        processors.append(structlog.processors.format_exc_info)
        # Set `severity`, not `level`, for compatibility with Google
        # Stackdriver logging expectations.
        processors.append(replace_level_with_severity)

    # Stackdriver logging expects a "message" field, not "event"
    processors.append(structlog.processors.EventRenamer("message"))

    structlog.configure(
        processors=processors
        + [structlog.stdlib.ProcessorFormatter.wrap_for_formatter],
        logger_factory=structlog.stdlib.LoggerFactory(),
        cache_logger_on_first_use=True,
    )

    if development_logs:
        log_renderer = structlog.dev.ConsoleRenderer(event_key="message")  # type: ignore
    else:
        log_renderer = structlog.processors.JSONRenderer()  # type: ignore

    formatter = structlog.stdlib.ProcessorFormatter(
        foreign_pre_chain=processors,
        processors=[
            structlog.stdlib.ProcessorFormatter.remove_processors_meta,
            log_renderer,
        ],
    )

    handler = logging.StreamHandler()
    handler.setFormatter(formatter)

    root = logging.getLogger()
    root.addHandler(handler)
    root.setLevel(log_level)

    # Propagate uvicorn logs instead of letting uvicorn configure the format
    for name in ["uvicorn", "uvicorn.access", "uvicorn.error"]:
        logging.getLogger(name).handlers.clear()
        logging.getLogger(name).propagate = True

    # Reconfigure log levels for some overly chatty libraries
    logging.getLogger("uvicorn.access").setLevel(logging.WARNING)
    logging.getLogger("urllib3.connectionpool").setLevel(logging.ERROR)



from typing import Iterator

from cog import BasePredictor, Path


class Predictor(BasePredictor):
    def predict(self, file: Path) -> Iterator[str]:
        with file.open() as f:
            prefix = f.read()
        predictions = ["foo", "bar", "baz"]
        for prediction in predictions:
            yield prefix + " " + prediction



import os
import uuid
from concurrent.futures import Future
from datetime import datetime
from unittest import mock

import pytest

from cog.schema import PredictionRequest, Status, WebhookEvent
from cog.server.eventtypes import Done, Log
from cog.server.runner import (
    PredictionRunner,
    PredictTask,
    RunnerBusyError,
    SetupResult,
    SetupTask,
    UnknownPredictionError,
)
from cog.server.worker import make_worker


def _fixture_path(name):
    test_dir = os.path.dirname(os.path.realpath(__file__))
    return os.path.join(test_dir, f"fixtures/{name}.py") + ":Predictor"


class FakeClock:
    def __init__(self, t):
        self.t = t

    def __call__(self):
        return self.t


tick = mock.sentinel.tick


class FakeWorker:
    def __init__(self):
        self.subscribers = {}
        self.subscribers_by_tag = {}

        self._setup_future = None
        self._predict_futures = {}
        self.last_prediction_payload = None

    def subscribe(self, subscriber, tag=None):
        sid = uuid.uuid4()
        self.subscribers[sid] = tag
        if tag not in self.subscribers_by_tag:
            self.subscribers_by_tag[tag] = {}
        self.subscribers_by_tag[tag][sid] = subscriber
        return sid

    def unsubscribe(self, sid):
        tag = self.subscribers.pop(sid)
        del self.subscribers_by_tag[tag][sid]

    def setup(self):
        assert self._setup_future is None
        self._setup_future = Future()
        return self._setup_future

    def run_setup(self, events):
        for event in events:
            if isinstance(event, Exception):
                self._setup_future.set_exception(event)
                return
            for subscriber in self.subscribers_by_tag.get(None, {}).values():
                subscriber(event)
            if isinstance(event, Done):
                self._setup_future.set_result(event)

    def predict(self, payload, tag=None):
        assert tag not in self._predict_futures or self._predict_futures[tag].done()
        self.last_prediction_payload = payload
        self._predict_futures[tag] = Future()
        print(f"setting {tag}, now {self._predict_futures}")
        return self._predict_futures[tag]

    def run_predict(self, events, id=None):
        if id is None:
            if len(self._predict_futures) != 1:
                raise ValueError("Could not guess prediction id, please specify")
            id = next(iter(self._predict_futures))
        for event in events:
            if isinstance(event, Exception):
                self._predict_futures[id].set_exception(event)
                return
            for subscriber in self.subscribers_by_tag.get(id, {}).values():
                subscriber(event)
            if isinstance(event, Done):
                print(f"reading {id} from {self._predict_futures}")
                self._predict_futures[id].set_result(event)

    def cancel(self, tag=None):
        done = Done(canceled=True)
        for subscriber in self.subscribers_by_tag.get(tag, {}).values():
            subscriber(done)
        self._predict_futures[tag].set_result(done)


def test_prediction_runner_setup_success():
    w = FakeWorker()
    r = PredictionRunner(worker=w)

    task = r.setup()
    assert not task.done()

    w.run_setup([Log(message="Setting up...", source="stdout")])
    assert task.result.logs == ["Setting up..."]
    assert not task.done()

    w.run_setup([Done()])
    assert task.done()
    assert task.result.status == Status.SUCCEEDED


def test_prediction_runner_setup_failure():
    w = FakeWorker()
    r = PredictionRunner(worker=w)
    task = r.setup()

    w.run_setup([Done(error=True)])
    assert task.done()
    assert task.result.status == Status.FAILED


def test_prediction_runner_setup_exception():
    w = FakeWorker()
    r = PredictionRunner(worker=w)

    task = r.setup()

    w.run_setup([RuntimeError("kaboom!")])
    assert task.done()
    assert task.result.status == Status.FAILED
    assert task.result.logs[0].startswith("Traceback")
    assert task.result.logs[0].endswith("kaboom!\n")


def test_prediction_runner_predict_success():
    w = FakeWorker()
    r = PredictionRunner(worker=w)

    r.setup()
    w.run_setup([Done()])

    task = r.predict(PredictionRequest(input={"text": "giraffes"}))
    assert w.last_prediction_payload == {"text": "giraffes"}
    assert task.result.input == {"text": "giraffes"}
    assert task.result.status == Status.PROCESSING

    w.run_predict([Log(message="Predicting...", source="stdout")])
    assert task.result.logs == "Predicting..."

    w.run_predict([Done()])
    assert task.result.status == Status.SUCCEEDED


def test_prediction_runner_predict_failure():
    w = FakeWorker()
    r = PredictionRunner(worker=w)

    r.setup()
    w.run_setup([Done()])

    task = r.predict(PredictionRequest(input={"text": "giraffes"}))
    assert w.last_prediction_payload == {"text": "giraffes"}
    assert task.result.input == {"text": "giraffes"}
    assert task.result.status == Status.PROCESSING

    w.run_predict([Done(error=True, error_detail="ErrNeckTooLong")])
    assert task.result.status == Status.FAILED
    assert task.result.error == "ErrNeckTooLong"


def test_prediction_runner_predict_exception():
    w = FakeWorker()
    r = PredictionRunner(worker=w)

    r.setup()
    w.run_setup([Done()])

    task = r.predict(PredictionRequest(input={"text": "giraffes"}))
    assert w.last_prediction_payload == {"text": "giraffes"}
    assert task.result.input == {"text": "giraffes"}
    assert task.result.status == Status.PROCESSING

    w.run_predict(
        [
            Log(message="counting shards\n", source="stdout"),
            Log(message="reticulating splines\n", source="stdout"),
            ValueError("splines not reticulable"),
        ]
    )

    assert task.result.logs.startswith("counting shards\nreticulating splines\n")
    assert "Traceback" in task.result.logs
    assert "ValueError: splines not reticulable" in task.result.logs
    assert task.result.status == Status.FAILED
    assert task.result.error == "splines not reticulable"


def test_prediction_runner_predict_before_setup():
    w = FakeWorker()
    r = PredictionRunner(worker=w)

    with pytest.raises(RunnerBusyError):
        r.predict(PredictionRequest(input={"text": "giraffes"}))


def test_prediction_runner_predict_before_setup_completes():
    w = FakeWorker()
    r = PredictionRunner(worker=w)

    r.setup()

    with pytest.raises(RunnerBusyError):
        r.predict(PredictionRequest(input={"text": "giraffes"}))


def test_prediction_runner_predict_before_predict_completes():
    w = FakeWorker()
    r = PredictionRunner(worker=w)

    r.setup()
    w.run_setup([Done()])

    r.predict(PredictionRequest(input={"text": "giraffes"}))

    with pytest.raises(RunnerBusyError):
        r.predict(PredictionRequest(input={"text": "giraffes"}))


def test_prediction_runner_predict_after_predict_completes():
    w = FakeWorker()
    r = PredictionRunner(worker=w)

    r.setup()
    w.run_setup([Done()])

    r.predict(PredictionRequest(id="p-1", input={"text": "giraffes"}))
    w.run_predict([Done()], id="p-1")

    r.predict(PredictionRequest(id="p-2", input={"text": "elephants"}))
    w.run_predict([Done()], id="p-2")

    assert w.last_prediction_payload == {"text": "elephants"}


def test_prediction_runner_is_busy():
    w = FakeWorker()
    r = PredictionRunner(worker=w)

    assert r.is_busy()

    r.setup()
    assert r.is_busy()

    w.run_setup([Done()])
    assert not r.is_busy()

    r.predict(PredictionRequest(input={"text": "elephants"}))
    assert r.is_busy()

    w.run_predict([Done()])
    assert not r.is_busy()


def test_prediction_runner_is_busy_concurrency():
    w = FakeWorker()
    r = PredictionRunner(worker=w, max_concurrency=3)

    assert r.is_busy()

    r.setup()
    assert r.is_busy()

    w.run_setup([Done()])
    assert not r.is_busy()

    r.predict(PredictionRequest(id="1", input={"text": "elephants"}))
    assert not r.is_busy()

    r.predict(PredictionRequest(id="2", input={"text": "elephants"}))
    assert not r.is_busy()

    r.predict(PredictionRequest(id="3", input={"text": "elephants"}))
    assert r.is_busy()

    w.run_predict([Done()], id="1")
    assert not r.is_busy()


def test_prediction_runner_predict_cancelation():
    w = FakeWorker()
    r = PredictionRunner(worker=w)

    r.setup()
    w.run_setup([Done()])

    task = r.predict(PredictionRequest(id="abcd1234", input={"text": "giraffes"}))

    with pytest.raises(ValueError):
        r.cancel(None)
    with pytest.raises(ValueError):
        r.cancel("")
    with pytest.raises(UnknownPredictionError):
        r.cancel("wxyz5678")

    w.run_predict([Log(message="Predicting...", source="stdout")])
    assert task.result.status == Status.PROCESSING

    r.cancel("abcd1234")
    assert task.result.status == Status.CANCELED


def test_prediction_runner_predict_cancelation_multiple_predictions():
    w = FakeWorker()
    r = PredictionRunner(worker=w)

    r.setup()
    w.run_setup([Done()])

    task1 = r.predict(PredictionRequest(id="abcd1234", input={"text": "giraffes"}))
    w.run_predict([Done()])

    task2 = r.predict(PredictionRequest(id="defg6789", input={"text": "elephants"}))
    with pytest.raises(UnknownPredictionError):
        r.cancel("abcd1234")

    r.cancel("defg6789")
    assert task1.result.status == Status.SUCCEEDED
    assert task2.result.status == Status.CANCELED


def test_prediction_runner_predict_cancelation_concurrent_predictions():
    w = FakeWorker()
    r = PredictionRunner(worker=w, max_concurrency=5)

    r.setup()
    w.run_setup([Done()])

    task1 = r.predict(PredictionRequest(id="abcd1234", input={"text": "giraffes"}))

    task2 = r.predict(PredictionRequest(id="defg6789", input={"text": "elephants"}))

    r.cancel("abcd1234")
    w.run_predict([Done()], id="defg6789")
    assert task1.result.status == Status.CANCELED
    assert task2.result.status == Status.SUCCEEDED


def test_prediction_runner_setup_e2e():
    w = make_worker(predictor_ref=_fixture_path("sleep"), is_async=False)
    r = PredictionRunner(worker=w)

    try:
        task = r.setup()
        task.wait(timeout=5)
    finally:
        w.shutdown()

    assert task.result.status == Status.SUCCEEDED
    assert task.result.logs == []
    assert isinstance(task.result.started_at, datetime)
    assert isinstance(task.result.completed_at, datetime)


def test_prediction_runner_predict_e2e():
    w = make_worker(predictor_ref=_fixture_path("sleep"), is_async=False)
    r = PredictionRunner(worker=w)

    try:
        r.setup().wait(timeout=5)
        task = r.predict(PredictionRequest(input={"sleep": 0.1}))
        task.wait(timeout=1)
    finally:
        w.shutdown()

    assert task.result.output == "done in 0.1 seconds"
    assert task.result.status == "succeeded"
    assert task.result.error is None
    assert task.result.logs == "starting\n"
    assert isinstance(task.result.started_at, datetime)
    assert isinstance(task.result.completed_at, datetime)


@pytest.mark.parametrize(
    "log,result",
    [
        (
            [],
            SetupResult(started_at=1),
        ),
        (
            [tick, Done()],
            SetupResult(started_at=1, completed_at=2, status=Status.SUCCEEDED),
        ),
        (
            [
                tick,
                Log("running 1\n", source="stdout"),
                Log("running 2\n", source="stdout"),
                Done(),
            ],
            SetupResult(
                started_at=1,
                completed_at=2,
                logs=["running 1\n", "running 2\n"],
                status=Status.SUCCEEDED,
            ),
        ),
        (
            [
                tick,
                tick,
                Done(error=True, error_detail="kaboom!"),
            ],
            SetupResult(
                started_at=1,
                completed_at=3,
                status=Status.FAILED,
            ),
        ),
    ],
)
def test_setup_task(log, result):
    c = FakeClock(t=1)
    t = SetupTask(_clock=c)

    for event in log:
        if event == tick:
            c.t += 1
        else:
            t.handle_event(event)

    assert t.result == result


def test_predict_task():
    p = PredictionRequest(
        input={"hello": "there"},
        id=None,
        created_at=None,
        output_file_prefix=None,
        webhook=None,
    )
    t = PredictTask(p)

    assert t.result.status == Status.PROCESSING
    assert t.result.output is None
    assert t.result.logs == ""
    assert isinstance(t.result.started_at, datetime)

    t.set_output_type(multi=False)
    t.append_output("giraffes")
    assert t.result.output == "giraffes"


def test_predict_task_multi():
    p = PredictionRequest(
        input={"hello": "there"},
        id=None,
        created_at=None,
        output_file_prefix=None,
        webhook=None,
    )
    t = PredictTask(p)

    assert t.result.status == Status.PROCESSING
    assert t.result.output is None
    assert t.result.logs == ""
    assert isinstance(t.result.started_at, datetime)

    t.set_output_type(multi=True)
    t.append_output("elephant")
    t.append_output("duck")
    assert t.result.output == ["elephant", "duck"]

    t.append_logs("running a prediction\n")
    t.append_logs("still running\n")
    assert t.result.logs == "running a prediction\nstill running\n"

    t.succeeded()
    assert t.result.status == Status.SUCCEEDED
    assert isinstance(t.result.completed_at, datetime)

    t.failed("oops")
    assert t.result.status == Status.FAILED
    assert t.result.error == "oops"
    assert isinstance(t.result.completed_at, datetime)

    t.canceled()
    assert t.result.status == Status.CANCELED
    assert isinstance(t.result.completed_at, datetime)


def test_predict_task_webhook_sender():
    p = PredictionRequest(
        input={"hello": "there"},
        id=None,
        created_at=None,
        output_file_prefix=None,
        webhook="https://a.url.honest",
    )
    t = PredictTask(p)
    t._webhook_sender = mock.Mock()
    t.track(Future())

    t._webhook_sender.assert_called_once_with(mock.ANY, WebhookEvent.START)
    actual = t._webhook_sender.call_args[0][0]
    assert actual.status == "processing"

    t.set_output_type(multi=True)
    t.append_output("elephant")
    t.append_output("duck")

    t.append_logs("running a prediction\n")
    t.append_logs("still running\n")

    t._webhook_sender.reset_mock()
    t.succeeded()

    t._webhook_sender.assert_called_once_with(
        mock.ANY,
        WebhookEvent.COMPLETED,
    )
    actual = t._webhook_sender.call_args[0][0]
    assert actual.input == {"hello": "there"}
    assert actual.output == ["elephant", "duck"]
    assert actual.logs == "running a prediction\nstill running\n"
    assert actual.status == "succeeded"
    assert "predict_time" in actual.metrics


def test_predict_task_webhook_sender_intermediate():
    p = PredictionRequest(
        input={"hello": "there"},
        id=None,
        created_at=None,
        output_file_prefix=None,
        webhook="https://a.url.honest",
    )
    t = PredictTask(p)
    t._webhook_sender = mock.Mock()
    t.track(Future())

    t._webhook_sender.assert_called_once_with(mock.ANY, WebhookEvent.START)
    actual = t._webhook_sender.call_args[0][0]
    assert actual.status == "processing"

    t._webhook_sender.reset_mock()
    t.set_output_type(multi=False)
    t.append_output("giraffes")
    assert t._webhook_sender.call_count == 0


def test_predict_task_webhook_sender_intermediate_multi():
    p = PredictionRequest(
        input={"hello": "there"},
        id=None,
        created_at=None,
        output_file_prefix=None,
        webhook="https://a.url.honest",
    )
    t = PredictTask(p)
    t._webhook_sender = mock.Mock()
    t.track(Future())

    t._webhook_sender.assert_called_once_with(mock.ANY, WebhookEvent.START)
    actual = t._webhook_sender.call_args[0][0]
    assert actual.status == "processing"

    t._webhook_sender.reset_mock()
    t.set_output_type(multi=True)
    t.append_output("elephant")
    print(t._webhook_sender.call_args_list)
    assert t._webhook_sender.call_count == 1
    actual = t._webhook_sender.call_args_list[0][0][0]
    assert actual.output == ["elephant"]
    assert t._webhook_sender.call_args_list[0][0][1] == WebhookEvent.OUTPUT

    t._webhook_sender.reset_mock()
    t.append_output("duck")
    assert t._webhook_sender.call_count == 1
    actual = t._webhook_sender.call_args_list[0][0][0]
    assert actual.output == ["elephant", "duck"]
    assert t._webhook_sender.call_args_list[0][0][1] == WebhookEvent.OUTPUT

    t._webhook_sender.reset_mock()
    t.append_logs("running a prediction\n")
    assert t._webhook_sender.call_count == 1
    actual = t._webhook_sender.call_args_list[0][0][0]
    assert actual.logs == "running a prediction\n"
    assert t._webhook_sender.call_args_list[0][0][1] == WebhookEvent.LOGS

    t._webhook_sender.reset_mock()
    t.append_logs("still running\n")
    assert t._webhook_sender.call_count == 1
    actual = t._webhook_sender.call_args_list[0][0][0]
    assert actual.logs == "running a prediction\nstill running\n"
    assert t._webhook_sender.call_args_list[0][0][1] == WebhookEvent.LOGS

    t._webhook_sender.reset_mock()
    t.succeeded()
    t._webhook_sender.assert_called_once()
    actual = t._webhook_sender.call_args[0][0]
    assert actual.status == "succeeded"
    assert t._webhook_sender.call_args[0][1] == WebhookEvent.COMPLETED

    t._webhook_sender.reset_mock()
    t.failed("oops")
    t._webhook_sender.assert_called_once()
    actual = t._webhook_sender.call_args[0][0]
    assert actual.status == "failed"
    assert actual.error == "oops"
    assert t._webhook_sender.call_args[0][1] == WebhookEvent.COMPLETED

    t._webhook_sender.reset_mock()
    t.canceled()
    t._webhook_sender.assert_called_once()
    actual = t._webhook_sender.call_args[0][0]
    assert actual.status == "canceled"
    assert t._webhook_sender.call_args[0][1] == WebhookEvent.COMPLETED


def test_predict_task_file_uploads():
    p = PredictionRequest(
        input={"hello": "there"},
        id=None,
        created_at=None,
        output_file_prefix=None,
        webhook=None,
    )
    t = PredictTask(p, upload_url="https://a.url.honest")
    t._file_uploader = mock.Mock()

    # in reality this would be a Path object, but in this test we just care it
    # passes the output into the upload files function and uses whatever comes
    # back as final output.
    t._file_uploader.return_value = "http://example.com/output-image.png"
    t.set_output_type(multi=False)
    t.append_output("Path(to/my/file)")

    t._file_uploader.assert_called_once_with("Path(to/my/file)")
    assert t.result.output == "http://example.com/output-image.png"


def test_predict_task_file_uploads_multi():
    p = PredictionRequest(
        input={"hello": "there"},
        id=None,
        created_at=None,
        output_file_prefix=None,
        webhook=None,
    )
    t = PredictTask(p, upload_url="https://a.url.honest")
    t._file_uploader = mock.Mock()

    t._file_uploader.return_value = []
    t.set_output_type(multi=True)

    t._file_uploader.return_value = "http://example.com/hello.jpg"
    t.append_output("hello.jpg")

    t._file_uploader.return_value = "http://example.com/world.jpg"
    t.append_output("world.jpg")

    t._file_uploader.assert_has_calls([mock.call("hello.jpg"), mock.call("world.jpg")])
    assert t.result.output == [
        "http://example.com/hello.jpg",
        "http://example.com/world.jpg",
    ]



import io
import pickle
import random
import string

import pytest
import responses

from cog.types import Secret, URLFile, URLPath, get_filename


def test_urlfile_protocol_validation():
    with pytest.raises(ValueError):
        URLFile("file:///etc/shadow")

    with pytest.raises(ValueError):
        URLFile("data:text/plain,hello")


def test_urlfile_custom_filename():
    u = URLFile("https://example.com/some-path", filename="my_file.txt")
    assert u.name == "my_file.txt"


@responses.activate
def test_urlfile_acts_like_response():
    responses.get(
        "https://example.com/some/url",
        json={"message": "hello world"},
        status=200,
    )

    u = URLFile("https://example.com/some/url")

    assert isinstance(u, io.IOBase)
    assert u.read() == b'{"message": "hello world"}'


@responses.activate
def test_urlfile_iterable():
    responses.get(
        "https://example.com/some/url",
        body="one\ntwo\nthree\n",
        status=200,
    )

    u = URLFile("https://example.com/some/url")
    result = list(u)

    assert result == [b"one\n", b"two\n", b"three\n"]


@responses.activate
def test_urlfile_no_request_if_not_used():
    # This test would be failed by responses if the request were actually made,
    # as we've not registered the handler for it.
    URLFile("https://example.com/some/url")


@responses.activate
def test_urlfile_can_be_pickled():
    u = URLFile("https://example.com/some/url")

    result = pickle.loads(pickle.dumps(u))

    assert isinstance(result, URLFile)


@responses.activate
def test_urlfile_can_be_pickled_even_once_loaded():
    mock = responses.get(
        "https://example.com/some/url",
        json={"message": "hello world"},
        status=200,
    )

    u = URLFile("https://example.com/some/url", filename="my_file.txt")
    assert u.name == "my_file.txt"
    assert u.read() == b'{"message": "hello world"}'

    result = pickle.loads(pickle.dumps(u))

    assert isinstance(result, URLFile)
    assert result.name == "my_file.txt"
    assert result.read() == b'{"message": "hello world"}'

    assert mock.call_count == 2


@pytest.mark.parametrize(
    "url,filename",
    [
        # Simple URLs
        ("https://example.com/test", "test"),
        ("https://example.com/test.jpg", "test.jpg"),
        (
            "https://example.com/ហ_ត_អ_វ_ប_នជ_ក_រស_គតរបស_ព_រ_យ_ស_ម_នអ_ណ_ចម_ល_Why_Was_The_Death_Of_Jesus_So_Powerful_.m4a",
            "ហ_ត_អ_វ_ប_នជ_ក_រស_គតរបស_ព_រ_យ_ស_ម_នអ_ណ_ចម_ល_Why_Was_The_Death_Of_Jesus_So_Powerful_.m4a",
        ),
        # Data URIs
        (
            "data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAUAAAAFCAYAAACNbyblAAAAHElEQVQI12P4//8/w38GIAXDIBKE0DHxgljNBAAO9TXL0Y4OHwAAAABJRU5ErkJggg==",
            "file.png",
        ),
        (
            "data:text/plain,hello world",
            "file.txt",
        ),
        (
            "data:application/data;base64,aGVsbG8gd29ybGQ=",
            "file",
        ),
        # URL-encoded filenames
        (
            "https://example.com/thing+with+spaces.m4a",
            "thing with spaces.m4a",
        ),
        (
            "https://example.com/thing%20with%20spaces.m4a",
            "thing with spaces.m4a",
        ),
        (
            "https://example.com/%E1%9E%A0_%E1%9E%8F_%E1%9E%A2_%E1%9E%9C_%E1%9E%94_%E1%9E%93%E1%9E%87_%E1%9E%80_%E1%9E%9A%E1%9E%9F_%E1%9E%82%E1%9E%8F%E1%9E%9A%E1%9E%94%E1%9E%9F_%E1%9E%96_%E1%9E%9A_%E1%9E%99_%E1%9E%9F_%E1%9E%98_%E1%9E%93%E1%9E%A2_%E1%9E%8E_%E1%9E%85%E1%9E%98_%E1%9E%9B_Why_Was_The_Death_Of_Jesus_So_Powerful_.m4a",
            "ហ_ត_អ_វ_ប_នជ_ក_រស_គតរបស_ព_រ_យ_ស_ម_នអ_ណ_ចម_ល_Why_Was_The_Death_Of_Jesus_So_Powerful_.m4a",
        ),
        # Illegal characters
        ("https://example.com/nulbytes\u0000.wav", "nulbytes_.wav"),
        ("https://example.com/nulbytes%00.wav", "nulbytes_.wav"),
        ("https://example.com/path%2Ftraversal.dat", "path_traversal.dat"),
        # Long filenames
        (
            "https://example.com/some/path/Biden_Trump_sows_chaos_makes_things_worse_U_S_hits_more_than_six_million_COVID_cases_WAPO_Trump_health_advisor_is_pushing_herd_immunity_strategy_despite_warnings_from_Fauci_medical_officials_Biden_says_he_hopes_to_be_able_to_visit_Wisconsin_as_governor_tells_Trump_to_stay_home_.mp3",
            "Biden_Trump_sows_chaos_makes_things_worse_U_S_hits_more_than_six_million_COVID_cases_WAPO_Trump_health_advisor_is_pushing_herd_immunity_strategy_despite_warnings_from_Fauci_medical_officials_Bide~.mp3",
        ),
        (
            "https://coppermerchants.example/complaints/𒀀𒈾𒂍𒀀𒈾𒍢𒅕𒆠𒉈𒈠𒌝𒈠𒈾𒀭𒉌𒈠𒀀𒉡𒌑𒈠𒋫𒀠𒇷𒆪𒆠𒀀𒄠𒋫𒀝𒁉𒄠𒌝𒈠𒀜𒋫𒀀𒈠𒄖𒁀𒊑𒁕𒄠𒆪𒁴𒀀𒈾𒄀𒅖𒀭𒂗𒍪𒀀𒈾𒀜𒁲𒅔𒋫𒀠𒇷𒅅𒈠𒋫𒀝𒁉𒀀𒄠.tablet",
            "𒀀𒈾𒂍𒀀𒈾𒍢𒅕𒆠𒉈𒈠𒌝𒈠𒈾𒀭𒉌𒈠𒀀𒉡𒌑𒈠𒋫𒀠𒇷𒆪𒆠𒀀𒄠𒋫𒀝𒁉𒄠𒌝𒈠𒀜𒋫𒀀𒈠𒄖𒁀𒊑𒁕𒄠𒆪𒁴𒀀𒈾𒄀𒅖~.tablet",
        ),
    ],
)
def test_get_filename(url, filename):
    assert get_filename(url) == filename


def test_secret_type():
    secret_value = "sw0rdf1$h"  # noqa: S105
    secret = Secret(secret_value)

    assert secret.get_secret_value() == secret_value
    assert str(secret) == "**********"


def test_truncate_filename_if_long():
    # Test that a file too long exception is not raised.
    random_str = "".join(
        random.SystemRandom().choice(string.ascii_uppercase + string.digits)
        for _ in range(350)
    )
    big_query = "query=" + random_str
    filename = "waffwyyg~.zip"
    url = "https://www.amazon.com/" + filename + "?" + big_query
    fileobj = io.BytesIO()
    url_path = URLPath(
        source=url,
        filename=filename + "?" + big_query,
        fileobj=fileobj,
    )
    assert url_path.filename == "waffwyyg~~.zip"
    _ = url_path.convert()



from cog import current_scope


class Predictor:
    def setup(self):
        print("did setup")

    async def predict(self, *, name: str) -> str:
        print(f"hello, {name}")

        current_scope().record_metric("foo", 123)

        return f"hello, {name}"



from cog import BaseModel, Input, Path

class TrainingOutput(BaseModel):
    weights: Path

def train(
    n: int,
) -> TrainingOutput:
    with open("weights.bin", "w") as fh:
        for _ in range(n):
            fh.write("a")

    return TrainingOutput(
        weights=Path("weights.bin"),
    )



from cog import BasePredictor, Path

from typing import List, Union


class Predictor(BasePredictor):
    def predict(self, args: Union[int, List[int]]) -> int:
        if isinstance(args, int):
            return args
        else:
            return sum(args)



from contextlib import contextmanager
from contextvars import ContextVar
from typing import Generator, Optional, TypedDict


# See: https://www.w3.org/TR/trace-context/
class TraceContext(TypedDict, total=False):  # pylint: disable=too-many-ancestors
    traceparent: str
    tracestate: str


TRACE_CONTEXT: ContextVar[Optional[TraceContext]] = ContextVar(
    "trace_context", default=None
)


def make_trace_context(
    traceparent: Optional[str] = None, tracestate: Optional[str] = None
) -> TraceContext:
    """
    Creates a trace context dictionary from the given traceparent and tracestate
    headers. This is used to pass the trace context between services.
    """
    ctx: TraceContext = {}
    if traceparent:
        ctx["traceparent"] = traceparent
    if tracestate:
        ctx["tracestate"] = tracestate
    return ctx


def current_trace_context() -> Optional[TraceContext]:
    """
    Returns the current trace context, this needs to be added via HTTP headers
    to all outgoing HTTP requests.
    """
    return TRACE_CONTEXT.get()


@contextmanager
def trace_context(ctx: TraceContext) -> Generator[None, None, None]:
    """
    A helper for managing the current trace context provided by the inbound
    HTTP request. This context is used to link requests across the system and
    needs to be added to all internal outgoing HTTP requests.
    """
    t = TRACE_CONTEXT.set(ctx)
    try:
        yield
    finally:
        TRACE_CONTEXT.reset(t)
